/*
 *			GPAC - MPEG-4 Systems C Development Kit
 *
 *			Copyright (c) Cyril Concolato 2005
 *					All rights reserved
 *
 *  This file is part of GPAC / command-line mp21 toolbox
 *
 *  GPAC is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *   
 *  GPAC is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *   
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 */

#include <gpac/intern/m4_isomedia_dev.h>
#include <stdio.h>

void DumpItemInfo(M21File *file, u32 itemID, Bool full_dump);
void DumpFileInfo(M21File *file);


void PrintVersion()
{
	fprintf(stdout, "MP21Box - GPAC version %s\n", M4_VERSION);
}

void PrintUsage()
{
	fprintf (stdout, "MP21Box [option] input [option]\n"
			"\nWritten by Cyril Concolato - ENST (c) 2005\n"
			"\n\t-m21: rewrites the input DID file as an MPEG-21 compliant file\n"
			"\n\t-res file: imports the resource file when creating the MPEG-21 compliant file (using -m21)\n"
			"\n\t-did: extracts the DID from the input MPEG-21 compliant file\n"
			"\n\t-out filename: specifies output file name. Default output is out_$INPUT\n"
			"\n\t-extract itemID file: extract the item into the specified file\n"
			"\n");
}


/*return value:
	0: not supported 
	1: ISO media 
	2: input XML file (.xml)
*/
u32 get_file_type_by_ext(char *inName)
{
	char *lowername;
	u32 type;
	lowername = strdup(inName);
	if (!lowername) return 0;
	strlwr(lowername);
	if (strstr(lowername, ".m21") || strstr(lowername, ".mp4") || strstr(lowername, ".3gp") || strstr(lowername, ".mov") || strstr(lowername, ".3g2")) type = 1;
	else if (strstr(lowername, ".xml")) type = 2;
	else type = 0;	
	free(lowername);
	return type;
}

int main(int argc, char **argv)
{
	char outfile[5000];
	M4Err e;
	char *arg, *inName, *outName, *mediaSource, *out_resource;
	u32 i;
	M21File *mp21;
	u8 m21mux, print_info;
	u32 info_item_id, extract_item_id;

	mp21 = NULL;
	inName = NULL;
	outName = NULL;
	mediaSource = NULL;

	m21mux = 0; 
	print_info = 0;
	info_item_id = 0;
	extract_item_id = 0;
	
	/*parse our args*/
	for (i = 1; i < (u32) argc ; i++) {
		arg = argv[i];
		if (arg[0] != '-') {
			inName = arg;
		}
		else if (!stricmp(arg, "-?")) { PrintUsage(); return 0; }
		else if (!stricmp(arg, "-version")) { PrintVersion(); return 0; }
		else if (!stricmp(arg, "-m21")) { m21mux = 1; }
		else if (!stricmp(arg, "-did")) { m21mux = 2; }
		else if (!stricmp(arg, "-res")) { mediaSource = argv[i+1]; i++;  }
		else if (!stricmp(arg, "-out")) { outName = argv[i+1]; i++; }
		else if (!stricmp(arg, "-extract")) { 
			extract_item_id = atoi(argv[i+1]);
			out_resource = strdup(argv[i+2]);
			i+=2;
		}
		else if (!stricmp(arg, "-info")) {
			print_info = 1;
			if ((i+1<(u32) argc) && (sscanf(argv[i+1], "%d", &info_item_id)==1)) {
				char szTk[20];
				sprintf(szTk, "%d", info_item_id);
				if (!strcmp(szTk, argv[i+1])) i++;
				else info_item_id=0;
			} else {
				info_item_id=0;
			}
		}

	}
	if (!inName) {
		PrintUsage();
		return 1;
	}
	e = M4OK;

	if (m21mux == 0) {
		switch (get_file_type_by_ext(inName)) {
		case 1:
			m21mux = 0;
			break;
		case 2:
			m21mux = 1;
			break;
		default:
			return M4NotSupported;
		}
	} 

	if (outName) {
		strcpy(outfile, outName);
	} else {
		strcpy(outfile, inName);
		while (outfile[strlen(outfile)-1] != '.') outfile[strlen(outfile)-1] = 0;
		outfile[strlen(outfile)-1] = 0;
		if (m21mux == 1)
			strcat(outfile, ".m21");
		else if (m21mux == 2)
			strcat(outfile, ".xml");
	}

	// Opening the file
	if (m21mux == 1) {
		mp21 = M21_FileOpen(outfile, M4_OPEN_WRITE);
	} else {
		mp21 = M21_FileOpen(inName, M4_OPEN_READ);
	}

	if (extract_item_id) {
		e = M21_ExtractHiddenItem(mp21, extract_item_id, out_resource);
		if (e) goto err_exit;
	}

	if (print_info) {
		if (info_item_id) DumpItemInfo(mp21, info_item_id, 1);
		else DumpFileInfo(mp21);
	}

	// Processing the file
	if (m21mux == 1) {		
		e = M21_AddDIDToXMLBox(mp21, inName);
		if (e) goto err_exit;

		if (mediaSource) {
			e = M21_AddHiddenResource(mp21, mediaSource);
			if (e) goto err_exit;
		}
	} else if (m21mux == 2) { // demuxing the MPEG-21 file
		e = M21_ExtractDIDFromXMLBox(mp21, outfile);
		if (e) goto err_exit;
	}
err_exit:
	M21_FileClose(mp21);
	return e;
}