/*
 *			GPAC - MPEG-4 Systems C Development Kit
 *
 *			Copyright (c) Jean Le Feuvre 2000-2005 
 *					All rights reserved
 *
 *  This file is part of GPAC / command-line mp4 toolbox
 *
 *  GPAC is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *   
 *  GPAC is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *   
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 */


#include <gpac/m4_author.h>
#include <gpac/m4_bifs.h>
#include <gpac/m4_scenegraph.h>


#ifndef M4_READ_ONLY

extern u32 swf_flags;
extern Float swf_flatten_angle;


void convert_file_info(char *inName, u32 trackID)
{
	char szType[5];
	M4Err e;
	u32 i;
	Bool found;
	M4TrackImporter import;
	memset(&import, 0, sizeof(M4TrackImporter));
	import.in_name = inName;
	import.flags = M4TI_PROBE_ONLY;
	e = MP4T_ImportMedia(&import);
	if (e) {
		fprintf(stdout, "Error probing file %s: %s\n", inName, M4ErrToString(e));
		return;
	}
	if (trackID) {
		fprintf(stdout, "Import probing results for track %s#%d:\n", inName, trackID);
	} else {
		fprintf(stdout, "Import probing results for %s:\n", inName);
		if (!import.nb_tracks) {
			fprintf(stdout, "File has no selectable tracks\n");
			return;
		}
		fprintf(stdout, "File has %d tracks\n", import.nb_tracks);
	}
	found = 0;
	for (i=0; i<import.nb_tracks; i++) {
		if (trackID && (trackID != import.tk_info[i].track_num)) continue;
		if (!trackID) fprintf(stdout, "\tTrack %d type: ", import.tk_info[i].track_num);
		else fprintf(stdout, "Track type: ");

		switch (import.tk_info[i].type) {
		case M4_VisualMediaType: fprintf(stdout, "Video\n"); break;
		case M4_AudioMediaType: fprintf(stdout, "Audio\n"); break;
		case M4_TimedTextMediaType: fprintf(stdout, "Text\n"); break;
		case M4_BIFSMediaType: fprintf(stdout, "BIFS\n"); break;
		default: fprintf(stdout, "Other (4CC: %s)\n", FCC_TO_STR(import.tk_info[i].type, szType)); break;
		}
		if (!trackID) continue;
		fprintf(stdout, "Import Capabilities:\n");
		if (import.tk_info[i].flags & M4TI_USE_DATAREF) fprintf(stdout, "\tCan use data referencing\n");
		if (import.tk_info[i].flags & M4TI_NO_FRAME_DROP) fprintf(stdout, "\tCan use fixed FPS import\n");
		if (import.tk_info[i].flags & M4TI_OVERRIDE_FPS) fprintf(stdout, "\tCan override source frame rate\n");
		if (import.tk_info[i].flags & (M4TI_SBR_IMPLICIT|M4TI_SBR_EXPLICIT)) fprintf(stdout, "\tCan use AAC-SBR signaling\n");
		if (import.tk_info[i].flags & M4TI_FORCE_MPEG4) fprintf(stdout, "\tCan force MPEG-4 Systems stream description\n");
		if (import.tk_info[i].flags & M4TI_3GPP_AGGREGATION) fprintf(stdout, "\tCan use 3GPP frame aggregation\n");
		if (import.tk_info[i].flags & M4TI_NO_DURATION) fprintf(stdout, "\tCannot use duration-based import\n");
		found = 1;
		break;
	}
	fprintf(stdout, "\n");
	if (!found && trackID) fprintf(stdout, "Cannot find track %d in file\n", trackID);
}

M4Err import_file(M4File *dest, char *inName, u32 import_flags, Float force_fps, u32 frames_per_sample)
{
	u32 track_id, i;
	Bool do_audio, do_video, do_all;
	M4Err e;
	M4TrackImporter import;
	char *ext, szName[1000];

	memset(&import, 0, sizeof(M4TrackImporter));

	strcpy(szName, inName);
	ext = strrchr(inName, '.');
	if (!ext) {
		fprintf(stdout, "Unknown input file type\n");
		return M4BadParam;
	}

	/*check duration import*/
	ext = strrchr(szName, '%');
	if (ext) {
		import.duration = atoi(ext+1) * 1000;
		ext[0] = 0;
	}
	
	/*select switches for av containers import*/
	do_audio = do_video = 0;
	track_id = 0;
	do_all = 1;
	ext = strrchr(szName, '#');
	if (ext) {
		if (!strnicmp(ext, "#audio", 6)) do_audio = 1;
		else if (!strnicmp(ext, "#video", 6)) do_video = 1;
		else if (!strnicmp(ext, "#trackID=", 9)) track_id = atoi(&ext[9]);
		else track_id = atoi(&ext[1]);
		ext[0] = 0;
	}
	if (do_audio || do_video || track_id) do_all = 0;

	import.in_name = szName;
	import.flags = M4TI_PROBE_ONLY;
	e = MP4T_ImportMedia(&import);
	if (e) return e;

	import.dest = dest;
	import.video_fps = force_fps;
	import.frames_per_sample = frames_per_sample;
	import.flags = import_flags;
	
	if (!import.nb_tracks) return MP4T_ImportMedia(&import);

	for (i=0; i<import.nb_tracks; i++) {
		import.trackID = import.tk_info[i].track_num;
		if (do_all) e = MP4T_ImportMedia(&import);
		else if (track_id && (track_id==import.trackID)) {
			track_id = 0;
			e = MP4T_ImportMedia(&import);
		}
		else if (do_audio && (import.tk_info[i].type==M4_AudioMediaType)) {
			do_audio = 0;
			e = MP4T_ImportMedia(&import);
		}
		else if (do_video && (import.tk_info[i].type==M4_VisualMediaType)) {
			do_video = 0;
			e = MP4T_ImportMedia(&import);
		}
		if (e) return e;
	}
	if (track_id) fprintf(stdout, "WARNING: Track ID %d not found in file\n", track_id);
	else if (do_video) fprintf(stdout, "WARNING: Video track not found\n");
	else if (do_audio) fprintf(stdout, "WARNING: Audio track not found\n");
	return e;
}


/*
		MPEG-4 encoding
*/

M4Err EncodeFile(char *in, M4File *mp4, char *logFile, char *mediaSource, u32 flags, u32 rap_freq) 
{
	M4Err e;
	M4ContextLoader load;
	M4SceneManager *ctx;
	LPSCENEGRAPH sg;
	
	sg = NewSceneGraph();
	ctx = NewSceneManager(sg);
	memset(&load, 0, sizeof(M4ContextLoader));
	load.fileName = in;
	load.ctx = ctx;
	load.cbk = "Parsing";
	load.OnProgress = MP4T_OnProgress;
	load.swf_import_flags = swf_flags;
	load.swf_flatten_limit = swf_flatten_angle;
	/*since we're encoding we must get MPEG4 nodes only*/
	load.flags = M4CL_MPEG4_STRICT;
	e = M4SM_LoaderInit(&load);
	if (e) {
		fprintf(stdout, "Cannot load context - %s\n", M4ErrToString(e));
		return e;
	}
	e = M4SM_LoaderRun(&load);
	M4SM_LoaderDone(&load);

	if (e) {
		fprintf(stdout, "Error loading file %s\n", M4ErrToString(e));
	} else {
		e = M4SM_EncodeFile(ctx, mp4, logFile, mediaSource, flags, rap_freq);
	}
	M4SM_Delete(ctx);
	SG_Delete(sg);

	M4_SetMovieVersionInfo(mp4, MP4_V2_File, 1);
	M4_ModifyAlternateBrand(mp4, ISO_Media_File, 1);
	return e;
}

/*
		MPEG-4 chunk encoding
*/

static u32 GetNbBits(u32 MaxVal)
{
	u32 k=0;
	while ((s32) MaxVal > ((1<<k)-1) ) k+=1;
	return k;
}

M4Err EncodeBIFSChunk(M4SceneManager *ctx, char *bifsOutputFile, M4Err (*AUCallback)(M4Sample *))
{
	M4Err			e;
	char *data;
	u32 data_len;
	LPBIFSENC		bifsenc;
	InitialObjectDescriptor *iod;
	u32				i, j, count;
	M4StreamContext *sc;
	ESDescriptor	*esd;
	Bool is_in_iod, delete_desc, encode_names;
	BIFSConfigDescriptor bcfg;
	M4AUContext		*au;
	M4Sample		*samp;
	M4File *dest;
	char szF[M4_MAX_PATH], *ext;

	strcpy(szF, bifsOutputFile);
	ext = strrchr(szF, '.');
	if (ext) ext[0] = 0;
	strcat(szF, ".mp4");


	/* step3: encoding all AUs in ctx->streams starting at AU index 1 (0 is SceneReplace from previous context) */
	/*
	from here, the coding is mainly the same as in M4SM_EncodeBIFS (see SceneManager.c)
	*/
	bifsenc = BIFS_NewEncoder(ctx->scene_graph);
	e = M4OK;

	iod = (InitialObjectDescriptor *) ctx->root_od;
	/*if no iod check we only have one bifs*/
	if (!iod) {
		count = 0;
		for (i=0; i<ChainGetCount(ctx->streams); i++) {
			sc = ChainGetEntry(ctx->streams, i);
			if (sc->streamType == M4ST_OD) count++;
		}
		if (!iod && count>1) return M4NotSupported;
	}

	dest = M4_MovieOpen(szF, M4_OPEN_WRITE);
	if (!dest) return M4_GetLastError(NULL);

	count = ChainGetCount(ctx->streams);

	for (i=0; i<ChainGetCount(ctx->streams); i++) {
		u32 track, di, nbb;
		M4StreamContext *sc = ChainGetEntry(ctx->streams, i);
		esd = NULL;
		if (sc->streamType != M4ST_SCENE) continue;

		delete_desc = 0;
		esd = NULL;
		is_in_iod = 1;
		if (iod) {
			is_in_iod = 0;
			for (j=0; j<ChainGetCount(iod->ESDescriptors); j++) {
				esd = ChainGetEntry(iod->ESDescriptors, j);
				if (esd->decoderConfig && esd->decoderConfig->streamType == M4ST_SCENE) {
					if (!sc->ESID) sc->ESID = esd->ESID;
					if (sc->ESID == esd->ESID) {
						is_in_iod = 1;
						break;
					}
				}
				/*special BIFS direct import from NHNT*/
				else if (ChainGetCount(iod->ESDescriptors)==1) {
					sc->ESID = esd->ESID;
					is_in_iod = 1;
					break;
				}
				esd = NULL;
			}
		}
		// if (!esd && sc->ESID) esd = M4SM_LocateESD(ctx, sc->ESID);

		if (!esd) {
			delete_desc = 1;
			esd = OD_NewESDescriptor(2);
			OD_DeleteDescriptor((Descriptor **) &esd->decoderConfig->decoderSpecificInfo);
			esd->ESID = sc->ESID;
			esd->decoderConfig->streamType = M4ST_SCENE;
		}

		/*should NOT happen (means inputctx is not properly setup)*/
		if (!esd->decoderConfig->decoderSpecificInfo) {
			memset(&bcfg, 0, sizeof(BIFSConfigDescriptor));
			bcfg.tag = BIFSConfig_Tag;
		}
		/*regular retrieve from ctx*/
		else if (esd->decoderConfig->decoderSpecificInfo->tag == BIFSConfig_Tag) {
			memcpy(&bcfg, (BIFSConfigDescriptor *)esd->decoderConfig->decoderSpecificInfo, sizeof(BIFSConfigDescriptor));
		}
		/*should not happen either (unless loading from MP4 in which case BIFSc is not decoded)*/
		else {
			OD_GetBIFSConfig(esd->decoderConfig->decoderSpecificInfo, esd->decoderConfig->objectTypeIndication, &bcfg);
		}
		/*NO CHANGE TO BIFSC otherwose the generated update will not match the input context*/
		nbb = GetNbBits(ctx->max_node_id);
		if (bcfg.nodeIDbits<nbb) fprintf(stdout, "Warning: BIFSConfig.NodeIDBits TOO SMALL\n");
		nbb = GetNbBits(ctx->max_route_id);
		if (bcfg.routeIDbits<nbb) fprintf(stdout, "Warning: BIFSConfig.RouteIDBits TOO SMALL\n");
		nbb = GetNbBits(ctx->max_proto_id);
		if (bcfg.protoIDbits<nbb) fprintf(stdout, "Warning: BIFSConfig.ProtoIDBits TOO SMALL\n");

		/*this is the real pb, not stored in cfg or file level, set at EACH replaceScene*/
		encode_names = 0;

		/* The BIFS Config that is passed here should be the BIFSConfig from the IOD */
		BIFS_NewStream(bifsenc, sc->ESID, &bcfg, encode_names, 0);

		/*setup MP4 track*/
		if (!esd->slConfig) esd->slConfig = (SLConfigDescriptor *) OD_NewDescriptor(SLConfigDescriptor_Tag);
		if (sc->timeScale) esd->slConfig->timestampResolution = sc->timeScale;
		if (!esd->slConfig->timestampResolution) esd->slConfig->timestampResolution = 1000;
		track = M4_NewTrack(dest, sc->ESID, M4_BIFSMediaType, esd->slConfig->timestampResolution );
		M4_SetTrackEnabled(dest, track, 1);
		if (!sc->ESID) sc->ESID = M4_GetTrackID(dest, track);
		esd->ESID = sc->ESID;
		BIFS_GetStreamConfig(bifsenc, sc->ESID, &data, &data_len);

		if (esd->decoderConfig->decoderSpecificInfo) OD_DeleteDescriptor((Descriptor **) &esd->decoderConfig->decoderSpecificInfo);
		esd->decoderConfig->decoderSpecificInfo = (DefaultDescriptor *) OD_NewDescriptor(DecoderSpecificInfo_Tag);
		esd->decoderConfig->decoderSpecificInfo->data = data;
		esd->decoderConfig->decoderSpecificInfo->dataLength = data_len;
		esd->decoderConfig->objectTypeIndication = BIFS_GetVersion(bifsenc, sc->ESID);		

		
		/*create stream description*/
		M4_NewStreamDescription(dest, track, esd, NULL, NULL, &di);
		if (is_in_iod) M4_AddTrackToRootOD(dest, track);

		for (j=0; j<ChainGetCount(sc->AUs); j++) {
			au = ChainGetEntry(sc->AUs, j);
			samp = M4_NewSample();
			/*in case using XMT*/
			if (au->timing_sec) au->timing = (u32) (au->timing_sec * esd->slConfig->timestampResolution);
			samp->DTS = au->timing;
			e = BIFS_EncodeAU(bifsenc, sc->ESID, au->commands, &samp->data, &samp->dataLength);
			if (AUCallback)
			{
				AUCallback(samp);
			} else {
				M4_AddSample(dest, track, 1, samp);
			}
		}
	}
	if (dest) {
		if (e) M4_MovieDelete(dest);
		else M4_MovieClose(dest);
	}
	return e;
}


/**
 * @chunkFile BT chunk to be encoded
 * @bifs output file name for the BIFS data
 * @inputContext initial BT upon which the chunk is based (shall not be NULL)
 * @outputContext: file name to dump the context after applying the new chunk to the input context 
                   can be NULL, without .bt
 * @logFile: can be NULL
 */
M4Err EncodeFileChunk(char *chunkFile, char *bifs, char *inputContext, char *outputContext) 
{
	M4Err			e;
	LPSCENEGRAPH	sg;
	M4SceneManager	*ctx;
	M4ContextLoader load;

	/*Step 1: create context and load input*/
	sg = NewSceneGraph();
	ctx = NewSceneManager(sg);
	memset(&load, 0, sizeof(M4ContextLoader));
	load.fileName = inputContext;
	load.ctx = ctx;
	/*since we're encoding we must get MPEG4 nodes only*/
	load.flags = M4CL_MPEG4_STRICT;
	e = M4SM_LoaderInit(&load);
	if (!e) e = M4SM_LoaderRun(&load);
	M4SM_LoaderDone(&load);
	if (e) {
		fprintf(stdout, "Cannot load context %s - %s\n", inputContext, M4ErrToString(e));
		goto exit;
	}

	/* Step 2: make sure we have only ONE RAP for each stream*/
	e = M4SM_MakeRandomAccess(ctx);
	if (e) goto exit;
	
	/*Step 3: loading the chunk into the context*/
	memset(&load, 0, sizeof(M4ContextLoader));
	load.fileName = chunkFile;
	load.ctx = ctx;
	load.flags = M4CL_MPEG4_STRICT | M4CL_CTX_IS_LOADED;
	e = M4SM_LoaderInit(&load);
	if (!e) e = M4SM_LoaderRun(&load);
	M4SM_LoaderDone(&load);
	if (e) {
		fprintf(stdout, "Cannot load chunk context %s - %s\n", chunkFile, M4ErrToString(e));
		goto exit;
	}

	/* Assumes that the first AU contains only one command a SceneReplace and 
	   that is not part of the current chunk */
	/* Last argument is a callback to pass the encoded AUs: not needed here 
	   Saving is not handled correctly */
	e = EncodeBIFSChunk(ctx, bifs, NULL); 
	if (e) goto exit;


	if (outputContext) {
		u32 d_mode, do_enc;
		char szF[M4_MAX_PATH], *ext;

		/*make random access for storage*/
		e = M4SM_MakeRandomAccess(ctx);
		if (e) goto exit;
		
		/*check if we dump to BT, XMT or encode to MP4*/
		strcpy(szF, outputContext);
		ext = strrchr(szF, '.');
		d_mode = M4SM_DUMP_BT;
		do_enc = 0;
		if (ext) {
			if (!stricmp(ext, ".xmt") || !stricmp(ext, ".xmta")) d_mode = M4SM_DUMP_XMTA;
			else if (!stricmp(ext, ".mp4")) do_enc = 1;
			ext[0] = 0;
		}

		if (do_enc) {
			M4File *mp4;
			strcat(szF, ".mp4");
			mp4 = M4_MovieOpen(szF, M4_WRITE_EDIT);
			e = M4SM_EncodeFile(ctx, mp4, NULL, NULL, 0, 0);
			if (e) M4_MovieDelete(mp4);
			else M4_MovieClose(mp4);
		}
		else e = M4SM_DumpToText(ctx, szF, d_mode);
	}

exit:
	if (ctx) {
		sg = ctx->scene_graph;
		M4SM_Delete(ctx);
		SG_Delete(sg);
	}
	return e;
}

#endif

