/*
 *			GPAC - MPEG-4 Systems C Development Kit
 *
 *			Copyright (c) Cyril Concolato 2004
 *					All rights reserved
 *
 *  This file is part of GPAC / SVG Scene Graph Generator sub-project
 *
 *  GPAC is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *   
 *  GPAC is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *   
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 */

#include <stdio.h>
#include <string.h>
#include <time.h>

#include <libxml/tree.h>
#include <libxml/parser.h>

#include <libxml/xpath.h>
#include <libxml/xpathInternals.h>

#include <gpac/m4_config.h>
#include <gpac/m4_tools.h>

#define RNG_NS "http://relaxng.org/ns/structure/1.0"
#define RNG_PREFIX "rng"

#define COPYRIGHT "/*\n *			GPAC - MPEG-4 Systems C Development Kit\n *\n *			Copyright (c) Cyril Concolato 2004\n *					All rights reserved\n *\n *  This file is part of GPAC / SVG Scene Graph sub-project\n *\n *  GPAC is free software; you can redistribute it and/or modify\n *  it under the terms of the GNU General Public License as published by\n *  the Free Software Foundation; either version 2, or (at your option)\n *  any later version.\n *\n *  GPAC is distributed in the hope that it will be useful,\n *  but WITHOUT ANY WARRANTY; without even the implied warranty of\n *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n *  GNU General Public License for more details.	\n *\n *  You should have received a copy of the GNU General Public License\n *  along with GNU Make; see the file COPYING.  If not, write to\n *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.\n *\n */\n"

typedef struct {
	xmlChar *svg_name;
	char impl_name[1000];

	xmlChar *svg_type;
	char impl_type[50];

	char default_value[100];
	Bool optional;
	Bool animatable;
	u8 inheritable; // 0 = false, 1 = true, 2 = explicit
} SVGAttribute;

typedef struct {
	xmlChar *svg_name;
	char impl_name[1000];
	Chain *attributes;
	Chain *attribute_refs;
} SVGAttributeRef;

typedef struct
{
	xmlChar *svg_name;
	char impl_name[1000];
	Chain *attributes;
	Chain *attribute_refs;
} SVGElement;

SVGAttributeRef *NewSVGAttributeRef()
{
	SVGAttributeRef *att_ref;
	SAFEALLOC(att_ref, sizeof(SVGAttributeRef))
	if (att_ref) {
		att_ref->attributes = NewChain();
		att_ref->attribute_refs = NewChain();
	}
	return att_ref;
}

SVGAttribute *NewSVGAttribute()
{
	SVGAttribute *att;
	SAFEALLOC(att, sizeof(SVGAttribute))
	return att;
}

SVGElement *NewSVGElement() 
{
	SVGElement *elt;
	SAFEALLOC(elt, sizeof(SVGElement));
	if (elt) {
		elt->attributes = NewChain();
		elt->attribute_refs = NewChain();
	}
	return elt;
}

void parseSVGAttributeStruct(xmlDocPtr doc, xmlXPathContextPtr xpathCtx, xmlNodePtr att_element, xmlChar *elt_svg_name, SVGAttribute *att);

void parseSVGAttributesRefStructSubElement(xmlDocPtr doc, xmlXPathContextPtr xpathCtx, xmlNodePtr node,
										   Chain *global_attribute_refs, 
										   Chain *local_attribute_refs, Chain *local_attributes, xmlChar *local_svg_name)
{
	xmlNodePtr att_node = node->xmlChildrenNode;
	while (att_node) {
		if (att_node->type == XML_ELEMENT_NODE) {
			if (!strcmp(att_node->name, "ref")) { /* The attribute is a list of attributes */

				SVGAttributeRef *att_ref = NULL;
				xmlChar *tmpRefName = xmlGetProp(att_node, "name");

				u32 i;
				for (i = 0; i < ChainGetCount(global_attribute_refs); i++) {
					SVGAttributeRef *temp_att_ref = (SVGAttributeRef *)ChainGetEntry(global_attribute_refs, i);
					if (!strcmp(tmpRefName, temp_att_ref->svg_name)) {
						att_ref = temp_att_ref;
						break;
					}
				}
				if (att_ref) {
					ChainAddEntry(local_attribute_refs, att_ref);
					ChainAddEntry(global_attribute_refs, att_ref);
				}
				xmlFree(tmpRefName);

			} else if (!strcmp(att_node->name, "optional")) { /* The attribute is optional */
				xmlNodePtr optionalsAtt = att_node->xmlChildrenNode;
				while (optionalsAtt) {
					if (optionalsAtt->type == XML_ELEMENT_NODE) {
						SVGAttribute *att = NewSVGAttribute();
						att->optional = 1;
						parseSVGAttributeStruct(doc, xpathCtx, optionalsAtt, local_svg_name, att);
						ChainAddEntry(local_attributes, att);
					}
					optionalsAtt = optionalsAtt->next;
				}
			} else if (!strcmp(att_node->name, "attribute")) { /* The attribute is required */
				xmlChar *animatable = xmlGetProp(att_node, "animatable");
				xmlChar *inheritable = xmlGetProp(att_node, "inheritable");
				SVGAttribute *att = NewSVGAttribute();
				att->optional = 0;
				if (animatable) att->animatable = !strcmp(animatable, "true");
				if (inheritable) {
					if (!strcmp(inheritable, "false")) {
						att->inheritable = 0;
					} else if (!strcmp(inheritable, "explicit")) {
						att->inheritable = 2;
					} else {
						att->inheritable = 1;
					}
				}
				parseSVGAttributeStruct(doc, xpathCtx, att_node, local_svg_name, att);
				ChainAddEntry(local_attributes, att);
				xmlFree(animatable);
				xmlFree(inheritable);
			} else {
				fprintf(stdout, "case not handled: attribute list has unknown child: %s in attributes list .\n", att_node->name);
			}
		}
		att_node = att_node->next;
	}

}

Chain* parseAllSVGAttributesRefStruct(xmlDocPtr doc, xmlXPathContextPtr xpathCtx)
{
	int i;
	Chain *global_attributes_ref = NULL;
	xmlXPathObjectPtr xpathObj; 
	xmlChar xpathExpr[100];
	sprintf(xpathExpr, "//rng:define[@name]");
	xpathObj = xmlXPathEvalExpression(xpathExpr, xpathCtx);				
	if (!xpathObj || xpathObj->type != XPATH_NODESET || !(xpathObj->nodesetval) || xpathObj->nodesetval->nodeNr == 0) return NULL;

	global_attributes_ref = NewChain();
	for (i = 0; i < xpathObj->nodesetval->nodeNr; i++) {
		xmlChar *tmp = NULL;
		xmlChar *define_name;
		xmlNodePtr node = xpathObj->nodesetval->nodeTab[i];

		if (node->type != XML_ELEMENT_NODE) continue;
		define_name = xmlGetProp(node, "name"); 
		if (strstr(define_name, ".attrib")) {
			SVGAttributeRef *ref = NULL;
		
			ref = NewSVGAttributeRef();
			ref->svg_name = define_name;
			strcpy(ref->impl_name, ref->svg_name);
			tmp = ref->impl_name;
			while ( (tmp = strstr(tmp, ".")) ) { *tmp='_'; tmp++; }
			tmp = ref->impl_name;
			while ( (tmp = strstr(tmp, "_")) ) { *tmp='_'; tmp++; }
			tmp = strstr(ref->impl_name, "attrib");
			if (tmp) {
				strcpy(tmp, "ATTRIB");
			} else {
				fprintf(stdout, "Warning: wrong naming in attrib ref: %s.\n", ref->svg_name);
				tmp = strstr(ref->impl_name, "attlist");
				if (tmp) strcpy(tmp, "ATTLIST");
			}
			tmp = NULL;

			parseSVGAttributesRefStructSubElement(doc, xpathCtx, node, global_attributes_ref, ref->attribute_refs, ref->attributes, ref->svg_name);

			ChainAddEntry(global_attributes_ref, ref);
		} else {
			xmlFree(define_name);
		}
	}
	xmlXPathFreeObject(xpathObj);

	return global_attributes_ref;
}

void parseSVGAttributeStruct(xmlDocPtr doc, xmlXPathContextPtr xpathCtx, xmlNodePtr att_element, xmlChar *elt_svg_name, SVGAttribute *att)
{
	xmlChar *tmp = NULL;
	xmlNodePtr tmpnode;
	if (!strcmp(elt_svg_name, "SVG.AnimationTimingNoMinMax.attrib")) {
		fprintf(stdout, "%s\n", elt_svg_name);
	}

	att->svg_name = xmlGetProp(att_element, "name");
	if (!strcmp(att->svg_name, "fill")) {
		if (strstr(elt_svg_name, "Animation")) {
			strcpy(att->impl_name, "freeze");
		} else {
			strcpy(att->impl_name, att->svg_name);
			tmp = att->impl_name;
			while ( (tmp = strstr(tmp, "-")) ) { *tmp='_'; tmp++; }
			tmp = att->impl_name;
			while ( (tmp = strstr(tmp, ":")) ) { *tmp='_'; tmp++; }
			tmp = NULL;
		}
	} else {
		strcpy(att->impl_name, att->svg_name);
		tmp = att->impl_name;
		while ( (tmp = strstr(tmp, "-")) ) { *tmp='_'; tmp++; }
		tmp = att->impl_name;
		while ( (tmp = strstr(tmp, ":")) ) { *tmp='_'; tmp++; }
		tmp = NULL;
	}

	tmpnode = att_element->xmlChildrenNode;
	while (tmpnode) {
		if (tmpnode->type == XML_ELEMENT_NODE) {
			if (!strcmp(tmpnode->name,"ref")) {
				att->svg_type = xmlGetProp(tmpnode, "name");
				sprintf(att->impl_type, "SVG_%s", att->svg_type);
				tmp = att->impl_type;
				while ( (tmp = strstr(tmp, ".")) ) { *tmp='_'; tmp++; }
				tmp = att->impl_type;;
				if ( (tmp = strstr(tmp, "datatype")) ) {
					tmp--;
					*tmp = 0;
				} else {
					fprintf(stdout, "Warning: type does not end with datatype: %s\n",att->svg_type);
				}
			}
		}
		tmpnode = tmpnode->next;
	}
	if (!att->svg_type) {
		if (!strcmp(att->svg_name, "visibility")) {
			strcpy(att->impl_type, "SVG_VisibilityValue");
		} else if (!strcmp(att->svg_name, "display")) {
			strcpy(att->impl_type, "SVG_DisplayValue");
		} else if (!strcmp(att->svg_name, "stroke-linecap")) {
			strcpy(att->impl_type, "SVG_StrokeLineCapValue");
		} else if (!strcmp(att->svg_name, "stroke-linejoin")) {
			strcpy(att->impl_type, "SVG_StrokeLineJoinValue");
		} else if (!strcmp(att->svg_name, "font-style")) {
			strcpy(att->impl_type, "SVG_FontStyleValue");
		} else if (!strcmp(att->impl_name, "freeze")) {
			strcpy(att->impl_type, "SMIL_FreezeValue");
		} else if (!strcmp(att->svg_name, "calcMode")) {
			strcpy(att->impl_type, "SMIL_CalcModeValue");
		} else if (!strcmp(att->svg_name, "values")) {
			strcpy(att->impl_type, "SMIL_AnimateValues");
		} else if (!strcmp(att->svg_name, "keyTimes")) {
			strcpy(att->impl_type, "SMIL_KeyTimesValues");
		} else if (!strcmp(att->svg_name, "keySplines")) {
			strcpy(att->impl_type, "SMIL_KeySplinesValues");
		} else if (!strcmp(att->svg_name, "from") || !strcmp(att->svg_name, "to") || !strcmp(att->svg_name, "by")) {
			strcpy(att->impl_type, "SMIL_AnimateValue");
		} else if (!strcmp(att->svg_name, "additive")) {
			strcpy(att->impl_type, "SMIL_AdditiveValue");
		} else if (!strcmp(att->svg_name, "accumulate")) {
			strcpy(att->impl_type, "SMIL_AccumulateValue");
		} else if (!strcmp(att->svg_name, "begin") ||
				   !strcmp(att->svg_name, "end")
				  ) {
			strcpy(att->impl_type, "SMIL_BeginOrEndValues");
		} else if (!strcmp(att->svg_name, "min") ||
				   !strcmp(att->svg_name, "max") ||
				   !strcmp(att->svg_name, "dur") ||
				   !strcmp(att->svg_name, "repeatDur")
				  ) {
			strcpy(att->impl_type, "SMIL_MinMaxDurRepeatDurValue");
		} else if (!strcmp(att->svg_name, "repeat")) {
			strcpy(att->impl_type, "SMIL_RepeatValue");
		} else if (!strcmp(att->svg_name, "restart")) {
			strcpy(att->impl_type, "SMIL_RestartValue");
		} else if (!strcmp(att->svg_name, "repeatCount")) {
			strcpy(att->impl_type, "SMIL_RepeatCountValue");
		} else if (!strcmp(att->svg_name, "attributeName")) {
			strcpy(att->impl_type, "SMIL_AttributeName");
		} else {
			strcpy(att->impl_type, "SVG_String");
			fprintf(stdout, "Warning: using SVG_String_datatype for attribute %s in element %s.\n", att->svg_name, elt_svg_name);
		}
	}
}

xmlNodePtr resolveRef(xmlDocPtr doc, xmlXPathContextPtr xpathCtx, xmlChar *refName)
{
	xmlXPathObjectPtr xpathObj; 
	xmlChar xpathExpr[100];
	sprintf(xpathExpr, "//rng:define[@name = \"%s\"]", refName);
	xpathObj = xmlXPathEvalExpression(xpathExpr, xpathCtx);				
	if (!xpathObj || xpathObj->type != XPATH_NODESET || !(xpathObj->nodesetval)) return NULL;
	if (xpathObj->nodesetval->nodeNr > 1) fprintf(stdout, "Warning: RNG Error, two define with the same name: %s.\n", refName);
	return xpathObj->nodesetval->nodeTab[0];
}

SVGElement *createSVGElementStruct(xmlDocPtr doc, xmlXPathContextPtr xpathCtx, xmlNodePtr svgnode, Chain *global_attribute_refs) 
{
	xmlChar *tmp;
	xmlNodePtr child;
	SVGElement *elt = NewSVGElement();

	elt->svg_name = xmlGetProp(svgnode, "name");
	if (!strcmp(elt->svg_name, "set")) {
		fprintf(stdout, "element: %s\n", elt->svg_name);
	}
	strcpy(elt->impl_name, elt->svg_name);

	tmp = elt->impl_name;
	while ( (tmp = strstr(tmp, "-")) ) { *tmp='_'; tmp++; }
	tmp = NULL;

	child = svgnode->xmlChildrenNode;
	while (child != NULL) {
		if (child->type == XML_ELEMENT_NODE) {
			if (!strcmp(child->name, "ref")) {
				xmlChar *childRefName;
				childRefName = xmlGetProp(child, "name");
				if (strstr(childRefName, "attlist.")) {
					/* List of attributes */
					xmlNodePtr derefed_attlist = resolveRef(doc, xpathCtx, childRefName);
					if (!derefed_attlist) continue;

					parseSVGAttributesRefStructSubElement(doc, xpathCtx, derefed_attlist, global_attribute_refs, elt->attribute_refs, elt->attributes, elt->svg_name);

					if (!strcmp(elt->svg_name, "text")) {
						SVGAttribute *string_att = NewSVGAttribute();
						SAFEALLOC(string_att->svg_name, sizeof(xmlChar *)*7);
						strcpy(string_att->svg_name, "string");
						strcpy(string_att->impl_name, "string");
						SAFEALLOC(string_att->svg_type, sizeof(xmlChar *)*14);
						strcpy(string_att->svg_type, "TextContent");
						strcpy(string_att->impl_type, "SVG_TextContent");
						ChainAddEntry(elt->attributes, string_att);
					}
				} else if (strstr(childRefName, ".content")) {
					/* Content model is not treated */
				} else {
					fprintf(stdout, "case not handled: element child is an unknown ref: %s in element .\n", childRefName);
				}
				xmlFree(childRefName);
			} else {
				fprintf(stdout, "case not handled: element child is not a ref.\n");
			}
		}
		child = child->next;
	}

	return elt;
}

void flattenAttributes(Chain *global_attribute_refs, Chain *global_attributes, Chain *local_attribute_refs, Chain *local_attributes)
{
	u32 i;
	for (i =0; i < ChainGetCount(local_attribute_refs); i++) {
		SVGAttributeRef *att_ref = ChainGetEntry(local_attribute_refs, i);
		flattenAttributes(global_attribute_refs, global_attributes, att_ref->attribute_refs, att_ref->attributes);
	}
	for (i = 0; i<ChainGetCount(local_attributes); i++) {
		SVGAttribute *att = (SVGAttribute *)ChainGetEntry(local_attributes, i);
		ChainAddEntry(global_attributes, att);
	}
}

void deleteAttributeRefs(Chain *attributes_refs)
{
	u32 i;
	for (i =0; i < ChainGetCount(attributes_refs); i++) {
		SVGAttributeRef *att_ref = ChainGetEntry(attributes_refs, i);
		deleteAttributeRefs(att_ref->attribute_refs);
		DeleteChain(att_ref->attribute_refs); att_ref->attribute_refs = NULL;
		DeleteChain(att_ref->attributes); att_ref->attributes = NULL;
	}
}

void deleteAllAttributeRefs(Chain *svg_elements)
{
	u32 i;
	for (i=0; i<ChainGetCount(svg_elements); i++) {
		SVGElement *elt = (SVGElement *)ChainGetEntry(svg_elements, i);
		deleteAttributeRefs(elt->attribute_refs);
		DeleteChain(elt->attribute_refs); elt->attribute_refs = NULL;
	}
}

void generateAttributes(FILE *output, Chain *attribute_refs, Chain *attributes) 
{
	u32 i;
	for (i = 0; i<ChainGetCount(attributes); i++) {
		SVGAttribute *att = (SVGAttribute *)ChainGetEntry(attributes, i);
		if (!strcmp(att->svg_name, "stroke-linejoin")) {
			fprintf(output, "\tSVG_StrokeLineJoinValue %s; /* %s, animatable: %s, inheritable: %s */\n", att->impl_name, (att->optional?"optional":"required"), (att->animatable?"yes":"no"), (att->inheritable?(att->inheritable>1?"explicit":"true"):"false"));
		} else if (!strcmp(att->svg_name, "stroke-linecap")) {
			fprintf(output, "\tSVG_StrokeLineCapValue %s; /* %s, animatable: %s, inheritable: %s */\n", att->impl_name, (att->optional?"optional":"required"), (att->animatable?"yes":"no"), (att->inheritable?(att->inheritable>1?"explicit":"true"):"false"));
		} else {
			fprintf(output, "\t%s %s; /* %s, animatable: %s, inheritable: %s */\n", att->impl_type, att->impl_name, (att->optional?"optional":"required"), (att->animatable?"yes":"no"), (att->inheritable?(att->inheritable>1?"explicit":"true"):"false"));
		}
	}
}

void generateNode(FILE *output, SVGElement* svg_elt, Chain *global_attribute_refs) 
{
	if (!strcmp(svg_elt->impl_name, "set")) {
		fprintf(stdout, "%s\n", svg_elt->impl_name);
	}
	fprintf(output, "typedef struct _tagSVG%sElement\n{\n", svg_elt->impl_name);
	fprintf(output, "\tBASE_NODE\n");
	fprintf(output, "\tCHILDREN\n");
	fprintf(output, "\tBASE_SVG_ELEMENT\n");
	generateAttributes(output, svg_elt->attribute_refs, svg_elt->attributes);
	fprintf(output, "} SVG%sElement;\n\n\n", svg_elt->impl_name);
}

static Chain *styling_property_names;

void generateNodeImpl(FILE *output, SVGElement* svg_elt) 
{
	u32 i;
	
	fprintf(output, "static void SVG_%s_Del(SFNode *node)\n{\n", svg_elt->impl_name);
	fprintf(output, "\tSVG%sElement *p = (SVG%sElement *)node;\n", svg_elt->impl_name, svg_elt->impl_name);
		for (i = 0; i < ChainGetCount(svg_elt->attributes); i++) {
		SVGAttribute *att = ChainGetEntry(svg_elt->attributes, i);
		if (!strcmp("SVG_TransformList", att->impl_type) ||
			!strcmp("SVG_Points", att->impl_type) || 
			!strcmp("SVG_Coordinates", att->impl_type) ||
			!strcmp("SMIL_KeyTimesValues", att->impl_type) ||
			!strcmp("SMIL_KeySplinesValues", att->impl_type) ||			
			!strcmp("SMIL_BeginOrEndValues", att->impl_type) 
			) {
			fprintf(output, "\tDeleteChain(p->%s);\n", att->impl_name);
		} else if (!strcmp("SVG_PathData", att->impl_type)) {
			fprintf(output, "\tDeleteChain(p->d.path_commands);\n");
			fprintf(output, "\tDeleteChain(p->d.path_points);\n");
		} else if (!strcmp("SMIL_AnimateValue", att->impl_type)) {
			fprintf(output, "\tfree(p->%s.value);\n",att->impl_name);
		} else if (!strcmp("SMIL_AnimateValues", att->impl_type)) {
			fprintf(output, "\tDeleteChain(p->%s.values);\n",att->impl_name);
		}
	}
	fprintf(output, "\tDestroyChildrenNode((SFNode *) p);\n");
	fprintf(output, "\tSFNode_Delete((SFNode *)p);\n");
	fprintf(output, "}\n\n");
	
	fprintf(output, "static M4Err SVG_%s_get_attribute(SFNode *node, FieldInfo *info)\n{\n", svg_elt->impl_name);
	fprintf(output, "\tswitch (info->fieldIndex) {\n");
	for (i = 0; i < ChainGetCount(svg_elt->attributes); i++) {
		SVGAttribute *att = ChainGetEntry(svg_elt->attributes, i);
		fprintf(output, "\t\tcase %d:\n", i);
		fprintf(output, "\t\t\tinfo->name = \"%s\";\n", att->svg_name);
		fprintf(output, "\t\t\tinfo->fieldType = %s_datatype;\n", att->impl_type);
		fprintf(output, "\t\t\tinfo->far_ptr = & ((SVG%sElement *)node)->%s;\n", svg_elt->impl_name, att->impl_name);
		fprintf(output, "\t\t\treturn M4OK;\n");
	}
	fprintf(output, "\t\tdefault: return M4BadParam;\n\t}\n}\n\n");

	fprintf(output, "void *SVG_New_%s()\n{\n\tSVG%sElement *p;\n", svg_elt->impl_name,svg_elt->impl_name);
	fprintf(output, "\tSAFEALLOC(p, sizeof(SVG%sElement));\n\tif (!p) return NULL;\n\tNode_Setup((SFNode *)p, TAG_SVG_%s);\n\tSetupChildrenNode((SFNode *) p);\n",svg_elt->impl_name,svg_elt->impl_name);
	fprintf(output, "#ifdef NODE_USE_POINTERS\n");
	fprintf(output, "\t((SFNode *p)->sgprivate->name = \"%s\";\n", svg_elt->impl_name);
	fprintf(output, "\t((SFNode *p)->sgprivate->node_del = SVG_%s_Del;\n", svg_elt->impl_name);
	fprintf(output, "\t((SFNode *p)->sgprivate->get_field = SVG_%s_get_attribute;\n", svg_elt->impl_name);
	fprintf(output, "#endif\n");

	if (strcmp(svg_elt->svg_name, "video") && strcmp(svg_elt->svg_name, "font-face")) {
		for (i = 0; i < ChainGetCount(svg_elt->attributes); i++) {
			u32 j;
			SVGAttribute *att = ChainGetEntry(svg_elt->attributes, i);

			/* default values should be handled more properly, generically */
			if (!strcmp(att->impl_name, "fill")) {
				fprintf(output, "\tp->fill.paintType = SVG_PAINTTYPE_INHERIT;\n");
			} else if (!strcmp(att->svg_name, "fill-rule")) {
				fprintf(output, "\tp->fill_rule = SVGFillRule_inherit;\n");
			} else if (!strcmp(att->svg_name, "fill-opacity")) {
				fprintf(output, "\tp->fill_opacity.type = SVGFLOAT_INHERIT;\n");
			} else if (!strcmp(att->svg_name, "stroke")) {
				fprintf(output, "\tp->stroke.paintType = SVG_PAINTTYPE_INHERIT;\n");
			} else if (!strcmp(att->svg_name, "stroke-opacity")) {
				fprintf(output, "\tp->stroke_opacity.type = SVGFLOAT_INHERIT;\n");
			} else if (!strcmp(att->svg_name, "stroke-width")) {
				fprintf(output, "\tp->stroke_width.unitType = SVG_LENGTHTYPE_INHERIT;\n");
			} else if (!strcmp(att->svg_name, "stroke-linejoin")) {
				fprintf(output, "\tp->stroke_linejoin = SVGStrokeLineJoin_inherit;\n");
			} else if (!strcmp(att->svg_name, "stroke-linecap")) {
				fprintf(output, "\tp->stroke_linecap = SVGStrokeLineCap_inherit;\n");
			} else if (!strcmp(att->svg_name, "stroke-miterlimit")) {
				fprintf(output, "\tp->stroke_miterlimit.type = SVGFLOAT_INHERIT;\n");
			} else if (!strcmp(att->svg_name, "font-size")) {
				fprintf(output, "\tp->font_size.type = SVGFLOAT_INHERIT;\n");
			} else if (!strcmp(att->svg_name, "min")) {
				fprintf(output, "\tp->min.type = SMILMinMaxDurRepeatDur_clock_value;\n");
			} else if (!strcmp(att->svg_name, "repeatCount")) {
				fprintf(output, "\tp->repeatCount = 1;\n");
			} else if (!strcmp(att->svg_name, "repeatDur")) {
				fprintf(output, "\tp->repeatDur.clock_value = -1;\n");
			} else if (!strcmp(att->svg_name, "calcMode") && !strcmp(svg_elt->svg_name, "animateMotion")) {
				fprintf(output, "\tp->calcMode = SMILCalcMode_paced;\n");
			} else 
				/* Inialization of complex types */
			if ( !strcmp("SVG_TransformList", att->impl_type) ||
						!strcmp("SVG_Points", att->impl_type) || 
						!strcmp("SVG_Coordinates", att->impl_type) ||
						!strcmp("SMIL_KeyTimesValues", att->impl_type) ||
						!strcmp("SMIL_KeySplinesValues", att->impl_type) ||			
						!strcmp("SMIL_BeginOrEndValues", att->impl_type) 
					  ) {
				fprintf(output, "\tp->%s = NewChain();\n", att->impl_name);
			} else if (!strcmp("SVG_PathData", att->impl_type)) {
				fprintf(output, "\tp->d.path_commands = NewChain();\n");
				fprintf(output, "\tp->d.path_points = NewChain();\n");
			} else if (!strcmp("SMIL_AnimateValues", att->impl_type)) {
				fprintf(output, "\tp->%s.values = NewChain();\n",att->impl_name);
			}

			for (j = 0; j < ChainGetCount(styling_property_names); j++) {
				char *prop_name = ChainGetEntry(styling_property_names, j);
				if (!strcmp(prop_name, att->svg_name) && strcmp("freeze", att->impl_name)) {
					fprintf(output, "\tp->properties.%s = &(p->%s);\n", att->impl_name, att->impl_name);
					break;
				}
			}
		}
	}
	fprintf(output, "\treturn p;\n}\n\n");

}

/*type: 0: header, 1: source*/
static FILE *BeginFile(u32 type)
{
	FILE *f;

	char sPath[M4_MAX_PATH];

	if (!type) {
		sprintf(sPath, "..%c..%c..%cinclude%cgpac%cm4_svg_nodes.h", M4_PATH_SEPARATOR, M4_PATH_SEPARATOR, M4_PATH_SEPARATOR, M4_PATH_SEPARATOR, M4_PATH_SEPARATOR);
	} else {
		sprintf(sPath, "..%c..%c..%cM4Systems%cSceneGraph%csvg_nodes.c", M4_PATH_SEPARATOR, M4_PATH_SEPARATOR, M4_PATH_SEPARATOR, M4_PATH_SEPARATOR, M4_PATH_SEPARATOR);
	}
	
	f = fopen(sPath, "wt");
	fprintf(f, "%s\n", COPYRIGHT);

	{
		time_t rawtime;
		time(&rawtime);
		fprintf(f, "\n/*\n\tDO NOT MOFIFY - File generated on GMT %s\n\tBY SVGGen for GPAC Version %s\n*/\n\n", asctime(gmtime(&rawtime)), M4_VERSION);
	}

	if (!type) {
		fprintf(f, "#ifndef _M4_SVG_NODES_H\n");
		fprintf(f, "#define _M4_SVG_NODES_H\n\n");
		fprintf(f, "#ifdef __cplusplus\nextern \"C\" {\n#endif\n\n");
	}
	return f;
}

static void EndFile(FILE *f, u32 type)
{
	if (!type) {
		fprintf(f, "#ifdef __cplusplus\n}\n#endif\n\n");
		fprintf(f, "\n\n#endif\t\t/*_M4_SVG_NODES_H*/\n\n");
	}
	fclose(f);
}

Chain *parseAllSVGElements(xmlDocPtr doc, xmlXPathContextPtr xpathCtx, Chain *attribute_refs_list)
{
	int k;
	Chain *svg_elements = NULL;
	xmlChar *xpathExpr = "//rng:element[@name]";
	xmlNodeSetPtr nodes;
	xmlXPathObjectPtr xpathObj; 

    /* Get all the RNG elements */
    xpathObj = xmlXPathEvalExpression(xpathExpr, xpathCtx);
    if(xpathObj == NULL || xpathObj->type != XPATH_NODESET) return NULL;
	
	svg_elements = NewChain();
	nodes = xpathObj->nodesetval;	
		
	for (k = 0; k < nodes->nodeNr; k++)	{
		xmlNodePtr node = nodes->nodeTab[k];
		if (k == 1) continue; /* duplicated svg element in current RNG schema */
		if (node->type == XML_ELEMENT_NODE) {
			SVGElement *elt = createSVGElementStruct(doc, xpathCtx, node, attribute_refs_list);
			flattenAttributes(attribute_refs_list, elt->attributes, elt->attribute_refs, NULL);
			ChainAddEntry(svg_elements, elt);
		}
	}
	xmlXPathFreeObject(xpathObj);

	return svg_elements;
}

int main (int argc, char **argv)
{
	xmlDocPtr doc = NULL;
	xmlXPathContextPtr xpathCtx = NULL; 
	Chain *svg_elements = NULL;
	Chain *svg_attribute_lists = NULL;
	u32 i;

	FILE *output;

    xmlInitParser();
	LIBXML_TEST_VERSION

	doc = xmlParseFile(argv[1]);
    if (!doc) {
        printf("error: could not parse file %s\n", argv[1]);
		return -1;
    }

	xpathCtx = xmlXPathNewContext(doc);
    if(xpathCtx == NULL) {
        fprintf(stderr,"Error: unable to create new XPath context\n");
        xmlFreeDoc(doc); 
        return(-1);
    }
	xmlXPathRegisterNs(xpathCtx, RNG_PREFIX, RNG_NS);

	svg_attribute_lists = parseAllSVGAttributesRefStruct(doc, xpathCtx);
	svg_elements = parseAllSVGElements(doc, xpathCtx, svg_attribute_lists);
	deleteAllAttributeRefs(svg_elements);
	DeleteChain(svg_attribute_lists);

	styling_property_names = NewChain();
	ChainAddEntry(styling_property_names, "fill");
	ChainAddEntry(styling_property_names, "fill-rule");
	ChainAddEntry(styling_property_names, "fill-opacity");
	ChainAddEntry(styling_property_names, "stroke");
	ChainAddEntry(styling_property_names, "stroke-opacity");
	ChainAddEntry(styling_property_names, "stroke-width");
	ChainAddEntry(styling_property_names, "stroke-linecap");
	ChainAddEntry(styling_property_names, "stroke-linejoin");
	ChainAddEntry(styling_property_names, "stroke-miterlimit");
	ChainAddEntry(styling_property_names, "font-size");
	ChainAddEntry(styling_property_names, "font-family");
	ChainAddEntry(styling_property_names, "font-style");

	output = BeginFile(0);
	fprintf(output, "#include <gpac/m4_scenegraph_svg.h>\n\n\n");
	fprintf(output, "/* Definition of SVG element internal tags */\n");
	fprintf(output, "/* TAG names are made of \"TAG_SVG\" + SVG element name (with - replaced by _) */\n");
	/* write all tags */
	fprintf(output, "enum {\n");
	for (i=0; i<ChainGetCount(svg_elements); i++) {
		SVGElement *elt = (SVGElement *)ChainGetEntry(svg_elements, i);
		if (i == 0) 
			fprintf(output, "\tTAG_SVG_%s = TAG_RANGE_FIRST_SVG", elt->impl_name);
		else 
			fprintf(output, ",\n\tTAG_SVG_%s", elt->impl_name);
	}
	fprintf(output, ",\n\t/*undefined elements (when parsing) use this tag*/\n\tTAG_SVG_UndefinedElement\n};\n\n");
	fprintf(output, "/******************************************\n");
 	fprintf(output, "*   SVG Elements structure definitions    *\n");
 	fprintf(output, "*******************************************/\n");
	for (i=0; i<ChainGetCount(svg_elements); i++) {
		SVGElement *elt = (SVGElement *)ChainGetEntry(svg_elements, i);
		generateNode(output, elt, svg_attribute_lists);
	}
	fprintf(output, "/******************************************\n");
 	fprintf(output, "*  End SVG Elements structure definitions *\n");
 	fprintf(output, "*******************************************/\n");
	EndFile(output, 0);


	output = BeginFile(1);
	fprintf(output, "#include <gpac/m4_svg_nodes.h>\n\n");
	fprintf(output, "#ifndef M4_DISABLE_SVG\n\n");
	fprintf(output, "#include <gpac/intern/m4_scenegraph_dev.h>\n\n");
	for (i=0; i<ChainGetCount(svg_elements); i++) {
		SVGElement *elt = (SVGElement *)ChainGetEntry(svg_elements, i);
		generateNodeImpl(output, elt);
	}

	/* SVGElement *SVG_CreateNode(u32 ElementTag)*/
	fprintf(output, "SVGElement *SVG_CreateNode(u32 ElementTag)\n");
	fprintf(output, "{\n");
	fprintf(output, "\tswitch (ElementTag) {\n");
	for (i=0; i<ChainGetCount(svg_elements); i++) {
		SVGElement *elt = (SVGElement *)ChainGetEntry(svg_elements, i);
		fprintf(output, "\t\tcase TAG_SVG_%s: return SVG_New_%s();\n",elt->impl_name,elt->impl_name);
	}
	fprintf(output, "\t\tdefault: return NULL;\n\t}\n}\n\n");
	
	/* void SVGElement_Del(SVGElement *elt) */
	fprintf(output, "void SVGElement_Del(SVGElement *elt)\n{\n");
	fprintf(output, "\tSFNode *node = (SFNode *)elt;\n");
	fprintf(output, "\tswitch (node->sgprivate->tag) {\n");
	for (i=0; i<ChainGetCount(svg_elements); i++) {
		SVGElement *elt = (SVGElement *)ChainGetEntry(svg_elements, i);
		fprintf(output, "\t\tcase TAG_SVG_%s: SVG_%s_Del(node); return;\n", elt->impl_name, elt->impl_name);
	}
	fprintf(output, "\t\tdefault: return;\n\t}\n}\n\n");

	/* u32 SVG_GetAttributeCount(SVGElement *elt) */
	fprintf(output, "u32 SVG_GetAttributeCount(SFNode *node)\n{\n");
	fprintf(output, "\tswitch (node->sgprivate->tag) {\n");
	for (i=0; i<ChainGetCount(svg_elements); i++) {
		SVGElement *elt = (SVGElement *)ChainGetEntry(svg_elements, i);
		fprintf(output, "\t\tcase TAG_SVG_%s: return %i;\n", elt->impl_name, ChainGetCount(elt->attributes));
	}
	fprintf(output, "\t\tdefault: return 0;\n\t}\n}\n\n");
	
	/* M4Err SVG_GetAttributeInfo(SFNode *node, FieldInfo *info) */
	fprintf(output, "M4Err SVG_GetAttributeInfo(SFNode *node, FieldInfo *info)\n{\n");
	fprintf(output, "\tswitch (node->sgprivate->tag) {\n");
	for (i=0; i<ChainGetCount(svg_elements); i++) {
		SVGElement *elt = (SVGElement *)ChainGetEntry(svg_elements, i);
		fprintf(output, "\t\tcase TAG_SVG_%s: return SVG_%s_get_attribute(node, info);\n", elt->impl_name, elt->impl_name);
	}
	fprintf(output, "\t\tdefault: return M4BadParam;\n\t}\n}\n\n");

	/* u32 SVG_GetTagByName(const char *element_name) */
	fprintf(output, "u32 SVG_GetTagByName(const char *element_name)\n{\n\tif (!element_name) return TAG_UndefinedNode;\n");
	for (i=0; i<ChainGetCount(svg_elements); i++) {
		SVGElement *elt = (SVGElement *)ChainGetEntry(svg_elements, i);
		fprintf(output, "\tif (!stricmp(element_name, \"%s\")) return TAG_SVG_%s;\n", elt->svg_name, elt->impl_name);
	}
	fprintf(output, "\treturn TAG_UndefinedNode;\n}\n\n");


	/* const char *SVG_GetElementName(u32 tag) */
	fprintf(output, "const char *SVG_GetElementName(u32 tag)\n{\n\tswitch(tag) {\n");
	for (i=0; i<ChainGetCount(svg_elements); i++) {
		SVGElement *elt = (SVGElement *)ChainGetEntry(svg_elements, i);
		fprintf(output, "\tcase TAG_SVG_%s: return \"%s\";\n", elt->impl_name, elt->svg_name);
	}
	fprintf(output, "\tdefault: return \"UndefinedNode\";\n\t}\n}\n\n");

	fprintf(output, "#endif /*M4_DISABLE_SVG*/\n\n");
	EndFile(output, 1);

	/* End of code generation */	
	/* Terminate program by freeing all */	
/*
	for (i=0; i<ChainGetCount(svg_elements); i++) {
		u32 j;
		SVGElement *elt = (SVGElement *)ChainGetEntry(svg_elements, i);
		for (j=0; j < ChainGetCount(elt->attributes); j++) {
			SVGAttribute *att = (SVGAttribute *)ChainGetEntry(elt->attributes, j);
			if (att->svg_type) {
				xmlFree(att->svg_type);
				att->svg_type = NULL;
			}
			xmlFree(att->svg_name);
			att->svg_name = NULL;
		}
		DeleteChain(elt->attributes);
		elt->attributes = NULL;
		xmlFree(elt->svg_name);
		elt->svg_name = NULL;
	}
	DeleteChain(svg_elements);
	svg_elements = NULL;
	DeleteChain(styling_property_names);
*/
	xmlXPathFreeContext(xpathCtx); 
	xmlFreeDoc(doc); 
	
	xmlCleanupParser();
	return 0;
}
