/*
 *			GPAC - MPEG-4 Systems C Development Kit
 *
 *			Copyright (c) Cyril Concolato 2005
 *					All rights reserved
 *
 *  This file is part of GPAC / ISO Media File Format sub-project
 *
 *  GPAC is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *   
 *  GPAC is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *   
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 */


#include <gpac/intern/m4_isomedia_dev.h>

//the only static var. Used to store any error happening while opening a movie
static M4Err MP21_API_IO_Err;

void M21SetLastError(M21File *m21, M4Err error)
{
	if (!m21) {
		MP21_API_IO_Err = error;
		return;
	}
	m21->LastError = error;
}


M4Err M21_GetLastError(M21File *m21)
{
	if (!m21) return MP21_API_IO_Err;
	return m21->LastError;
}

u8 M21_GetOpenMode(M21File *m21)
{
	if (!m21) return 0;
	return m21->openMode;
}

M21File *M21NewFile()
{
	M21File *file = (M21File*)malloc(sizeof(M21File));
	if (file == NULL) {
		M4SetLastError(NULL, M4OutOfMem);
		return NULL;
	}
	memset(file, 0, sizeof(M21File));

	//init the atoms
	file->TopAtoms = NewChain();
	if (!file->TopAtoms) {
		M4SetLastError(NULL, M4OutOfMem);
		free(file);
		return NULL;
	}
	return file;
}

void M21_DelFile(M21File *mp21)
{
	//these are our two main files
	if (mp21->fileMap) DataMap_Delete(mp21->fileMap);

#ifndef M4_READ_ONLY
	if (mp21->editFileMap) {
		DataMap_Delete(mp21->editFileMap);
	}
	if (mp21->finalName) free(mp21->finalName);
#endif	//M4_READ_ONLY

	if (mp21->brand) DelAtom((Atom *)mp21->brand);
	DeleteAtomList(mp21->TopAtoms);

	if (mp21->fileName) free(mp21->fileName);
	free(mp21);
}

/**************************************************************
					File Reading
**************************************************************/
M4Err ParseRootAtom(Atom **outAtom, BitStream *bs, u64 *bytesExpected);

M4Err ParseFileAtoms(M21File *mp21, u64 *bytesMissing)
{
	Atom *a;
	u64 totSize;
	M4Err e = M4OK;

	totSize = 0;


	//while we have some data, parse our atoms...
	while (BS_Available(mp21->fileMap->bs)) {
		*bytesMissing = 0;

		e = ParseRootAtom(&a, mp21->fileMap->bs, bytesMissing);

		if (e >= 0) {
			e = M4OK;
		} else if (e == M4UncompleteFile) {
			//our mdat is uncomplete, only valid for READ ONLY files...
			if (mp21->openMode != M4_OPEN_READ) {
				return M4InvalidMP4File;
			}
			return e;
		} else {
			return e;
		}

		switch (a->type) {
		//our MOOV atom
		case MetaAtomType:
			if (mp21->meta) return M4InvalidMP4File;
			mp21->meta = (MetaAtom *)a;
			//add it to our list
			e = ChainAddEntry(mp21->TopAtoms, a);
			if (e) return e;
			totSize += a->size;
			break;
		//we only keep the MDAT in READ for dump purposes
		case MediaDataAtomType:
			totSize += a->size;
			if (!mp21->mdat && mp21->openMode == M4_OPEN_READ) {
				mp21->mdat = (MediaDataAtom *) a;
				e = ChainAddEntry(mp21->TopAtoms, mp21->mdat);
				if (e) return e;
			}
			//if we don't have any MDAT yet, create one (edit-write mode)
			//We only work with one mdat, but we're puting it at the place
			//of the first mdat found when opening a file for editing
			else if (!mp21->mdat && mp21->openMode != M4_OPEN_READ) {
				DelAtom(a);
				mp21->mdat = (MediaDataAtom *) CreateAtom(MediaDataAtomType);
				e = ChainAddEntry(mp21->TopAtoms, mp21->mdat);
				if (e) return e;
			} else {
				DelAtom(a);
			}
			break;
		case FileTypeAtomType:
			if (mp21->brand) {
				//just ignore it
				DelAtom(a);
			} else {
				mp21->brand = (FileTypeAtom *)a;
				totSize += a->size;
			}
			break;

		default:
			totSize += a->size;
			e = ChainAddEntry(mp21->TopAtoms, a);
			break;
		}
	}
	
	if (!mp21->meta) return M4InvalidMP4File;

	if (mp21->openMode == M4_OPEN_READ) return M4OK;
	return M4OK;
}

#ifndef M4_READ_ONLY

#define MP21_CPRT_NOTICE "MP21 File Produced with GPAC"

M4Err M21_InsertCopyright(M21File *mp21)
{
	u32 i;
	Atom *a;
	FreeSpaceAtom *free;

	if (!mp21) return M4BadParam;
	//check if it's one of our demo files...
	for (i = 0; i < ChainGetCount(mp21->TopAtoms); i++) {
		a = ChainGetEntry(mp21->TopAtoms, i);
		if (a->type == FreeSpaceAtomType) {
			free = (FreeSpaceAtom *)a;
			if (free->dataSize)
				if (!strcmp(free->data, MP21_CPRT_NOTICE))	return M4OK;
		}
	}
	a = CreateAtom(FreeSpaceAtomType);
	if (!a) return M4OutOfMem;
	free = (FreeSpaceAtom *)a;
	free->dataSize = strlen(MP21_CPRT_NOTICE) + 1;
	free->data = malloc(free->dataSize);
	if (!free->data) return M4OutOfMem;
	strcpy(free->data, MP21_CPRT_NOTICE);
	return ChainAddEntry(mp21->TopAtoms, free);
}

typedef struct {
	u8 isDone;

	u32 extent_number;

	ItemLocationEntry *iloc_entry;
	ItemInfoEntry *iinf_entry;

	FILE *item_file;
	u64 item_file_length;
} ItemWriter;

void M21_CleanWriters(Chain *writers)
{
	ItemWriter *writer;
	while (ChainGetCount(writers)) {
		writer = (ItemWriter*)ChainGetEntry(writers, 0);
		fclose(writer->item_file);
		free(writer);
		ChainDeleteEntry(writers, 0);
	}
}

// Before calling this function, the full path of the item needs to be set.
M4Err M21_SetupWriters(M21File *mp21, Chain *writers)
{
	u32 i, item_count;
	ItemWriter *writer;

	if (!mp21->meta->item_locations) return M4OK;

	item_count = ChainGetCount(mp21->meta->item_locations->location_entries);
	for (i = 0; i < item_count; i++) {
		writer = (ItemWriter*)malloc(sizeof(ItemWriter));
		if (!writer) goto exit;
		memset(writer, 0, sizeof(ItemWriter));
		writer->iloc_entry = ChainGetEntry(mp21->meta->item_locations->location_entries, i);
		writer->iinf_entry = ChainGetEntry(mp21->meta->item_infos->item_infos, i);
		if (!writer->iinf_entry->full_path) goto exit;
		writer->item_file = f64_open(writer->iinf_entry->full_path, "rb");
		if (!writer->item_file) goto exit;
		f64_seek(writer->item_file, 0, SEEK_END);
		writer->item_file_length = f64_tell(writer->item_file);
		f64_seek(writer->item_file, 0, SEEK_SET);
		ChainAddEntry(writers, writer);	
	}
	return M4OK;
exit:
	M21_CleanWriters(writers);
	return M4OutOfMem;
}

static u32 GetNbBits(u64 MaxVal)
{
	u32 k=0;
	while ((s32) MaxVal > ((1<<k)-1) ) k+=1;
	return k;
}

// No interleaving, so all items are written in one extent
M4Err M21_WriteAllItemsInOneExtent(M21File *mp21, Chain *writers, BitStream *bs, u8 Emulation, u64 StartOffset)
{
	u32 i, count;
	u64 maxExtentLength = 0, maxBaseOffset = 0;
	ItemWriter *writer;

	count = ChainGetCount(writers);
	for (i = 0; i < count; i++) {
		writer = (ItemWriter*)ChainGetEntry(writers, i);
		while (!writer->isDone) {
			ItemExtentEntry *entry = malloc(sizeof(ItemExtentEntry));
			if (!entry) return M4OutOfMem;
			memset(entry, 0, sizeof(ItemExtentEntry));

			if (writer->iloc_entry->data_reference_index) {
				// The content of item is left outside of the file 
				// We create one extent entry and set it to 0
			} else {
				// The content of the item is actually written in this file
		
				// we need to set the base offset of all the extents for this item
				// == start position of the first entry in the current file
				writer->iloc_entry->base_offset = BS_GetPosition(bs);
				// the following test should always be true
				if (writer->iloc_entry->base_offset > maxBaseOffset) maxBaseOffset = writer->iloc_entry->base_offset;

				// we need to set entry_length and offset for the only extent
				// entry_offset based on the base_offset for this item == always 0 because one extent.
				entry->extent_offset = 0;

				// entry_length == item_length
				
				entry->extent_length = writer->item_file_length;
				if (!Emulation) {
					char cache_data[4096];
					u64 remain = entry->extent_length;
					while (remain) {
						u32 size_cache = (remain>4096) ? 4096 : (u32) remain;
						fread(cache_data, 1, size_cache, writer->item_file);
						BS_WriteData(bs, cache_data, size_cache);
						remain -= size_cache;
					}
				}
				if (entry->extent_length > maxExtentLength) maxExtentLength = entry->extent_length;
			}
			ChainAddEntry(writer->iloc_entry->extent_entries, entry);
			writer->extent_number++;
			writer->isDone = 1;
		}
	}
	// Determine offset_size, length_size and base_offset_size of ItemLocationAtom
	mp21->meta->item_locations->offset_size = 0; // since all extent_offset are set to 0
	
	// To do: compute the nbbits for maxBaseOffset(ie. the last base_offset) and round to 0, 4 or 8.
	mp21->meta->item_locations->base_offset_size = GetNbBits(maxBaseOffset); 
	if (mp21->meta->item_locations->base_offset_size > 8*4) mp21->meta->item_locations->base_offset_size = 8;
	if (mp21->meta->item_locations->base_offset_size > 0) mp21->meta->item_locations->base_offset_size = 4;
	
	// To do: compute the nbbits for maxextentlength and round to 0, 4 or 8.
	mp21->meta->item_locations->length_size = GetNbBits(maxExtentLength); 
	if (mp21->meta->item_locations->length_size > 8*4) mp21->meta->item_locations->length_size = 8;
	if (mp21->meta->item_locations->length_size > 0) mp21->meta->item_locations->length_size = 4;

	return M4OK;
}

//write item per item, with meta atom before or after the mdat
M4Err M21_WriteFlat(M21File *mp21, u8 metaFirst, BitStream *bs)
{
	M4Err e;
	u32 i;
	u64 mdat_totSize, mdat_begin, offset;
	Atom *a;
	Chain *writers = NewChain();

	mdat_begin = mdat_totSize = 0;

	//first setup the writers
	e = M21_SetupWriters(mp21, writers);
	if (e) goto exit;

/*
	//the brand atom is always first no matter what
	if (mp21->brand) {
		e = SizeAtom((Atom *)mp21->brand);
		if (e) goto exit;
		e = WriteAtom((Atom *)mp21->brand, bs);
		if (e) goto exit;
	}
*/
	for (i = 0; i < ChainGetCount(mp21->TopAtoms); i++) {
		a = (Atom*)ChainGetEntry(mp21->TopAtoms, i);
		switch (a->type) {
		case MetaAtomType:
			// Do not yet write the meta atom, wait for the mdat to be written
			break;
		case MediaDataAtomType:
			mdat_begin = BS_GetPosition(bs);
			BS_WriteInt(bs, 0, 64);
			BS_WriteInt(bs, 0, 64);
			e = M21_WriteAllItemsInOneExtent(mp21, writers, bs, 0, BS_GetPosition(bs));
			if (e) goto exit;
			mdat_totSize = BS_GetPosition(bs) - mdat_begin;
			break;
		default:
			e = SizeAtom(a);
			if (e) goto exit;
			e = WriteAtom(a, bs);
			if (e) goto exit;
		}
	}

	//OK, write the meta atom.
	e = SizeAtom((Atom *)mp21->meta);
	if (e) return e;
	e = WriteAtom((Atom *)mp21->meta, bs);
	if (e) goto exit;

	if (mp21->mdat) {
		offset = BS_GetPosition(bs);
		e = BS_Seek(bs, mdat_begin);
		if (e) goto exit;
		//size
		if (mdat_totSize > 0xFFFFFFFF) {
			BS_WriteInt(bs, 1, 32);
		} else {
			BS_WriteInt(bs, (u32) mdat_totSize, 32);
		}
		//type
		BS_WriteInt(bs, MediaDataAtomType, 32);
		//extendedSize
		if (mdat_totSize > 0xFFFFFFFF) {
			BS_WriteLongInt(bs, mdat_totSize, 64);
		}
		e = BS_Seek(bs, offset);
		mp21->mdat->size = mdat_totSize;
	}
	goto exit;

exit:
	M21_CleanWriters(writers);
	DeleteChain(writers);
	return e;
}


M4Err M21_WriteToFile(M21File *mp21)
{
	FILE *stream;
	BitStream *bs;
	M4Err e = M4OK;
	if (!mp21) return M4BadParam;

	if (mp21->openMode == M4_OPEN_READ) return M4BadParam;

	e = M21_InsertCopyright(mp21);
	if (e) return e;

	if (mp21->openMode == M4_OPEN_WRITE) {
		//capture mode: we don't need a new bitstream
		e = M21_WriteFlat(mp21, 0, mp21->editFileMap->bs);
	} else {
		//OK, we need a new bitstream
		stream = f64_open(mp21->finalName, "wb");
		if (!stream) return M4IOErr;
		bs = NewBitStreamFromFile(stream, BS_FILE_WRITE_ONLY);
		if (!bs) {
			fclose(stream);
			return M4OutOfMem;
		}

		e = M21_WriteFlat(mp21, 0, bs);	
		DeleteBitStream(bs);
		fclose(stream);
	}
	return e;
}


//update file type in WRITE mode
static M4Err M21_UpdateFileType(M21File *m21)
{
	M4Err e;
	if (m21->openMode != M4_OPEN_WRITE) return M4OK;

	//rewind
	BS_Seek(m21->editFileMap->bs, 0);
	m21->editFileMap->curPos = 0;

	//store brand
	e = SizeAtom((Atom *)m21->brand);
	if (e) return e;
	e = WriteAtom((Atom *)m21->brand, m21->editFileMap->bs);
	if (e) return e;

	return M4OK;
}

static M4Err M21_CanAccessFile(M21File *m21, u32 Mode)
{
	if (!m21) return M4BadParam;
	if (m21->openMode < Mode) return M4InvalidMP4Mode;
	return M4OK;
}

static M4Err M21_SetFileVersionInfo(M21File *m21, u32 MajorBrand, u32 MinorVersion)
{
	M4Err e;

	// Check if the major brand is correct
	if (!MajorBrand) return M4BadParam;

	// check if the file can be written otherwise the brand cannot be set.
	e = M21_CanAccessFile(m21, M4_OPEN_WRITE);
	if (e) return e;

	// No brand currently present in the file, we need to create one.
	if (!m21->brand) m21->brand = (FileTypeAtom *) CreateAtom(FileTypeAtomType);
	
	// if the brand was present or not, we forces the new major brand and version
	m21->brand->majorBrand = MajorBrand;
	m21->brand->minorVersion = MinorVersion;

	// if no alternate brand, we create one
	if (!m21->brand->altBrand) {
		m21->brand->altBrand = malloc(sizeof(u32));
		m21->brand->altBrand[0] = MajorBrand;
		m21->brand->altCount = 1;
		return M21_UpdateFileType(m21);
	} else { // otherwise we add the current one.
		u32 i, *p;

		//if the alternate brand was already present don't change anything
		for (i=0; i<m21->brand->altCount; i++) {
			if (m21->brand->altBrand[i] == MajorBrand) return M4OK;
		}

		// otherwise create a new alternate brand
		p = malloc(sizeof(u32)*(m21->brand->altCount + 1));
		if (!p) return M4OutOfMem;
		memcpy(p, m21->brand->altBrand, sizeof(u32)*m21->brand->altCount);
		p[m21->brand->altCount] = MajorBrand;
		m21->brand->altCount += 1;
		free(m21->brand->altBrand);
		m21->brand->altBrand = p;
		//rewrite begining of the file
		return M21_UpdateFileType(m21);
	}
}

static M4Err M21_ModifyAlternateBrand(M21File *m21, u32 Brand, u8 AddIt)
{
	u32 i, k, *p;
	M4Err e;

	if (!Brand || !m21->brand) return M4BadParam;

	e = M21_CanAccessFile(m21, M4_OPEN_WRITE);
	if (e) return e;
	
	//do not mofify major one
	if (!AddIt && m21->brand->majorBrand == Brand) return M4OK;

	if (!AddIt && m21->brand->altCount == 1) {
		//fixes it in case
		m21->brand->altBrand[0] = m21->brand->majorBrand;
		return M4OK;
	}
	
	//check for the brand
	for (i=0; i<m21->brand->altCount; i++) {
		if (m21->brand->altBrand[i] == Brand) goto found;
	}
	
	//Not found
	if (!AddIt) return M4OK;
	
	//add it
	p = malloc(sizeof(u32)*(m21->brand->altCount + 1));
	if (!p) return M4OutOfMem;
	memcpy(p, m21->brand->altBrand, sizeof(u32)*m21->brand->altCount);
	p[m21->brand->altCount] = Brand;
	m21->brand->altCount += 1;
	free(m21->brand->altBrand);
	m21->brand->altBrand = p;
	
	//rewrite begining
	return M21_UpdateFileType(m21);

found:
	//found
	if (AddIt) return M4OK;
	assert(m21->brand->altCount>1);

	//remove it
	p = malloc(sizeof(u32)*(m21->brand->altCount - 1));
	if (!p) return M4OutOfMem;
	k = 0;
	for (i=0; i<m21->brand->altCount; i++) {
		if (m21->brand->altBrand[i] == Brand) continue;
		else {
			p[k] = m21->brand->altBrand[i];
			k++;
		}
	}
	m21->brand->altCount -= 1;
	free(m21->brand->altBrand);
	m21->brand->altBrand = p;
	
	//rewrite begining
	return M21_UpdateFileType(m21);
}

//Create an MPEG-21 File for WRITE only (M4_OPEN_EDIT, M4_WRITE_EDIT)
static M21File *M21FileCreate(const char *fileName, u8 OpenMode)
{
	M4Err e;

	M21File *mp21 = M21NewFile();
	if (!mp21) return NULL;
	mp21->openMode = OpenMode;

	//in WRITE, the input dataMap is ALWAYS NULL
	mp21->fileMap = NULL;

	//but we have the edit one
	if (OpenMode == M4_OPEN_WRITE) {
		//this file will be the final file...
		mp21->fileName = strdup(fileName);
		e = DataMap_New(fileName, NULL, DM_MODE_W, & mp21->editFileMap);
		if (e) goto err_exit;
		
		M21_SetFileVersionInfo(mp21, MP21_File, 1);
		M21_ModifyAlternateBrand(mp21, ISO_Media_File, 1);

	} else { //OpenMode == M4_OPEN_EDIT == M4_WRITE_EDIT
		//we are in EDIT mode but we are creating the file -> temp file
		mp21->finalName = (char*)malloc(strlen(fileName) + 1);
		strcpy(mp21->finalName, fileName);
		e = DataMap_New("mp21_tmp_edit", NULL, DM_MODE_W, & mp21->editFileMap);
		if (e) goto err_exit;

		M21_SetFileVersionInfo(mp21, MP21_File, 1);
		M21_ModifyAlternateBrand(mp21, ISO_Media_File, 1);
	}

	//OK, create our atoms 
	mp21->meta = (MetaAtom *) meta_New();
	ChainAddEntry(mp21->TopAtoms, mp21->meta);

	mp21->meta->handler = (HandlerAtom *)hdlr_New();
	mp21->meta->handler->handlerType = M4_MPEG21MediaType;
	mp21->meta->handler->nameLength = strlen("GPAC MPEG-21 Handler") + 1;
	mp21->meta->handler->nameUTF8 = strdup("GPAC MPEG-21 Handler");

	//default behaviour is capture mode, no interleaving (eg, no rewrite of mdat)
	mp21->storageMode = M4_FLAT;
	return mp21;

err_exit:
	M21SetLastError(NULL, e);
	if (mp21) M21_DelFile(mp21);
	return NULL;
}
#endif

//Create and parse the movie for M4_OPEN_EDIT - M4_OPEN_READ only
static M21File *M21FileOpen(const char *fileName, u8 OpenMode)
{
	M4Err e;
	u64 bytes;
	M21File *mp21 = M21NewFile();
	if (!mp21) return NULL;

	mp21->fileName = strdup(fileName);
	mp21->openMode = OpenMode;

	if (OpenMode == M4_OPEN_READ) {
		e = DataMap_New(fileName, NULL, DM_MODE_RO, &mp21->fileMap);
		if (e) goto err_exit;
	} else { // OpenMode == M4_OPEN_EDIT
#ifdef M4_READ_ONLY
		//not allowed for READ_ONLY lib
		e = M4InvalidMP4Mode;
		goto err_exit;
#else
		//set a default output name for edited file
		mp21->finalName = (char*)malloc(strlen(fileName) + 5);
		if (!mp21->finalName) {
			e = M4OutOfMem;
			goto err_exit;
		}

		strcpy(mp21->finalName, "out_");
		strcat(mp21->finalName, fileName);

		//open the original file with edit tag
		e = DataMap_New(fileName, NULL, DM_MODE_E, &mp21->fileMap);
		//if the file doesn't exist, we assume it's wanted and create one from scratch
		if (e) goto err_exit;

		//and create a temp fileName for the edit
		e = DataMap_New("mp21_tmp_edit", NULL, DM_MODE_W, & mp21->editFileMap);
		if (e) goto err_exit;

#endif
	}

	//OK, let's parse the file...
	e = ParseFileAtoms(mp21, &bytes);
	if (e) goto err_exit;
	return mp21;
err_exit:
	M4SetLastError(NULL, e);
	M21_DelFile(mp21);
	return NULL;
}

M4Err M21_FileClose(M21File *mp21)
{
	M4Err e;
	if (mp21 == NULL) return M4InvalidMP4File;
	e = M4OK;
#ifndef M4_READ_ONLY
	//write the file
	if (mp21->openMode != M4_OPEN_READ) {
		e = M21_WriteToFile(mp21);
	}
#endif	//M4_READ_ONLY
	//free and return;
	M21_DelFile(mp21);
	return e;
}

M21File *M21_FileOpen(const char *fileName, u8 OpenMode)
{
	M21File *m21;

	MP21_API_IO_Err = M4OK;

	
	switch (OpenMode) {
//	case M4_OPEN_READ_DUMP:
	case M4_OPEN_READ:
		m21 = M21FileOpen(fileName, OpenMode);
		break;

#ifndef M4_READ_ONLY
	case M4_OPEN_WRITE:
		m21 = M21FileCreate(fileName, OpenMode);
		break;
	case M4_OPEN_EDIT:
		m21 = M21FileOpen(fileName, OpenMode);
		break;
	case M4_WRITE_EDIT:
		OpenMode = M4_OPEN_EDIT;
		m21 = M21FileCreate(fileName, OpenMode);
		break;
#endif	//M4_READ_ONLY
		
	default:
		return NULL;
	}
	return m21;
}

M4Err M21_AddDIDToXMLBox(M21File *m21, char *inName)
{
	u32 read;
	FILE *didfile;
	XMLAtom *xml;
	u32 xml_length;

	xml = (XMLAtom *)xml_New();
	if (!xml) return M4OutOfMem;
	ChainAddEntry(m21->meta->other_boxes, xml);

	/*assume 32bit max size*/
	didfile = fopen(inName, "rt");
	if (!didfile) return M4BadParam;
	fseek(didfile, 0, SEEK_END);
	xml_length = ftell(didfile) + 1;
	fseek(didfile, 0, SEEK_SET);
	xml->xml = malloc(sizeof(unsigned char)*xml_length);
	read = fread(xml->xml, 1, sizeof(unsigned char)*xml_length,  didfile);
	if (read == 0) {
		free(xml->xml);
		return M4BadParam;
	}
	fclose(didfile);
	xml->xml[read] = 0;
	return M4OK;
}

M4Err M21_ExtractDIDFromXMLBox(M21File *m21, char *outName)
{
	u32 i, count;
	XMLAtom *xml = NULL;
	FILE *didfile;

	// Find XMLBox
	count = ChainGetCount(m21->meta->other_boxes);
	for (i = 0; i <count; i++) {
		Atom *a = ChainGetEntry(m21->meta->other_boxes, i);
		if (a->type == XMLAtomType) {
			xml = (XMLAtom *)a;
			break;
		}
	}
	if (!xml || !xml->xml) return M4BadParam;
	
	didfile = f64_open(outName, "wt");
	if (!didfile) return M4BadParam;
	fwrite(xml->xml, 1, strlen(xml->xml)+1, didfile);
	fclose(didfile);
	return M4OK;
}

M4Err M21_AddHiddenResource(M21File *m21, char *resource)
{
	u32 lastItemID = 0;
	ItemInfoEntry *infe;

	if (m21->meta->item_infos) {
		u32 i;
		u32 item_count = ChainGetCount(m21->meta->item_infos->item_infos);
		for (i = 0; i < item_count; i++) {
			ItemInfoEntry *e= ChainGetEntry(m21->meta->item_infos->item_infos, i);
			if (e->item_ID > lastItemID) lastItemID = e->item_ID;
		}
	}

	{ // Creation of the ItemInfoEntry based on the mediaSource
		infe = (ItemInfoEntry *)infe_New();
		infe->item_ID = ++lastItemID;
		infe->item_name = strdup(resource);
		infe->full_path = strdup(resource);
		{
			char *extension = &(resource[strlen(resource) - 1]);
			while (*extension != '.' && extension != resource) extension--;
			if (!stricmp(extension, ".jpg")) {
				infe->content_type = strdup("image/jpg");
			} else if (!stricmp(extension, ".mp4")) {
				infe->content_type = strdup("video/mp4");
			} else {
				infe_del((Atom *)infe);
				infe = NULL;
			}
		}
	}

	if (infe) { // Creation of the ItemLocation 
		ItemLocationEntry *location_entry = malloc(sizeof(ItemLocationEntry));
		if (location_entry) {
			memset(location_entry, 0, sizeof(ItemLocationEntry));
			location_entry->extent_entries = NewChain();
			
			// Creates an mdat if it does not exist
			if (!m21->mdat) {
				m21->mdat = (MediaDataAtom *)mdat_New();
				ChainAddEntry(m21->TopAtoms, m21->mdat);
			}

			// Creation an ItemLocation Box if it does not exist
			if (!m21->meta->item_locations) m21->meta->item_locations = (ItemLocationAtom *)iloc_New();

			// Adds the current item location entry
			ChainAddEntry(m21->meta->item_locations->location_entries, location_entry);
			
			location_entry->item_ID = lastItemID;
			location_entry->data_reference_index = 0; // means the current file

		}

		if (!m21->meta->item_infos) m21->meta->item_infos = (ItemInfoAtom *)iinf_New();
		ChainAddEntry(m21->meta->item_infos->item_infos, infe);
	}
	return M4OK;
}

M4Err M21_ExtractHiddenResources(M21File *m21)
{
	FILE *resource = NULL;
	u32 i, item_count, j, location_count, k, extent_count;
	if (!m21 || !m21->meta || !m21->meta->item_infos || !m21->meta->item_locations) return M4BadParam;

	item_count = ChainGetCount(m21->meta->item_infos->item_infos);
	for (i = 0; i < item_count; i++) {
		ItemInfoEntry *item_entry= ChainGetEntry(m21->meta->item_infos->item_infos, i);
		ItemLocationEntry *location_entry = NULL;
		location_count = ChainGetCount(m21->meta->item_locations->location_entries);
		for (j= 0; j < location_count; j++) {
			ItemLocationEntry *tmp = ChainGetEntry(m21->meta->item_locations->location_entries, j);
			if (tmp->item_ID == item_entry->item_ID) {
				location_entry = tmp;
				break;
			}
		}

		if (!location_entry || location_entry->data_reference_index) continue;

		resource = f64_open(item_entry->item_name, "wb");
		
		extent_count = ChainGetCount(location_entry->extent_entries);
		for (k = 0; k < extent_count; k++) {
			char buf_cache[4096];
			u64 remain;
			ItemExtentEntry *extent_entry = ChainGetEntry(location_entry->extent_entries, k);
			BS_Seek(m21->fileMap->bs, location_entry->base_offset + extent_entry->extent_offset);

			remain = extent_entry->extent_length;
			while (remain) {
				u32 cache_size = (remain>4096) ? 4096 : (u32) remain;
				BS_ReadData(m21->fileMap->bs, buf_cache, cache_size);
				fwrite(buf_cache, 1, cache_size, resource);
				remain -= cache_size;
			}
		}
		fclose(resource);
	}

	return M4OK;
}

M4Err M21_ExtractHiddenItem(M21File *m21, u32 itemID, char *out_resource)
{
	FILE *resource = NULL;
	u32 i, item_count, j, location_count, k, extent_count;
	ItemInfoEntry *item_entry = NULL;
	ItemLocationEntry *location_entry = NULL;

	if (!m21 || !m21->meta || !m21->meta->item_infos || !m21->meta->item_locations) return M4BadParam;

	item_count = ChainGetCount(m21->meta->item_infos->item_infos);
	for (i = 0; i < item_count; i++) {
		ItemInfoEntry *tmp_info = ChainGetEntry(m21->meta->item_infos->item_infos, i);
		if (tmp_info->item_ID == itemID) {
			item_entry = tmp_info;
		}
	}
	location_count = ChainGetCount(m21->meta->item_locations->location_entries);
	for (j= 0; j < location_count; j++) {
		ItemLocationEntry *tmp_location = ChainGetEntry(m21->meta->item_locations->location_entries, j);
		if (tmp_location->item_ID == itemID) {
			location_entry = tmp_location;
			break;
		}
	}

	if (!item_entry || !location_entry) return M4BadParam;
	if (location_entry->data_reference_index) {
		return M4BadParam;
	}

	resource = f64_open(out_resource, "wb");
		
	extent_count = ChainGetCount(location_entry->extent_entries);
	for (k = 0; k < extent_count; k++) {
		char buf_cache[4096];
		u64 remain;
		ItemExtentEntry *extent_entry = ChainGetEntry(location_entry->extent_entries, k);
		BS_Seek(m21->fileMap->bs, location_entry->base_offset + extent_entry->extent_offset);


		remain = extent_entry->extent_length;
		while (remain) {
			u32 cache_size = (remain>4096) ? 4096 : (u32) remain;
			BS_ReadData(m21->fileMap->bs, buf_cache, cache_size);
			fwrite(buf_cache, 1, cache_size, resource);
			remain -= cache_size;
		}
	}
	fclose(resource);
	return M4OK;
}

u32 M21_GetItemCount(M21File *m21)
{
	if (!m21 || !m21->meta || !m21->meta->item_infos || !m21->meta->item_locations) return 0;
	return ChainGetCount(m21->meta->item_infos->item_infos);
}

M4Err M21_GetFileVersionInfo(M21File *m21, u32 *brand, u32 *minorVersion, u32 *AlternateBrandsCount)
{
	if (!m21 || !brand) return M4BadParam;
	if (!m21->brand) {
		*brand = 0;
		if (minorVersion) *minorVersion = 0;
		if (AlternateBrandsCount) *AlternateBrandsCount = 0;
		return M4OK;
	}

	*brand = m21->brand->majorBrand;
	if (minorVersion) *minorVersion = m21->brand->minorVersion;
	if (AlternateBrandsCount) *AlternateBrandsCount = m21->brand->altCount;
	return M4OK;

}

