/*
 *			GPAC - MPEG-4 Systems C Development Kit
 *
 *			Copyright (c) Jean Le Feuvre 2000-2004 
 *					All rights reserved
 *
 *  This file is part of GPAC / 3D rendering plugin
 *
 *  GPAC is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *   
 *  GPAC is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *   
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 */



#include "render3d_nodes.h"

typedef struct
{
	SoundInterface snd_ifce;
	SFVec3f pos;
} Sound2DStack;

static void DestroySound2D(SFNode *node)
{
	Sound2DStack *st = (Sound2DStack *)Node_GetPrivate(node);
	free(st);
}

/*sound2D wraper - spacialization is not supported yet*/
static void RenderSound2D(SFNode *node, void *rs)
{
	RenderEffect3D *eff = (RenderEffect3D*) rs;
	M_Sound2D *snd = (M_Sound2D *)node;
	Sound2DStack *st = (Sound2DStack *)Node_GetPrivate(node);

	if (!snd->source) return;

	/*this implies no DEF/USE for real location...*/
	st->pos.x = snd->location.x;
	st->pos.y = snd->location.y;
	st->pos.z = 0;
	mx_apply_vec(&eff->model_matrix, &st->pos);

	eff->sound_holder = &st->snd_ifce;
	Node_Render((SFNode *) snd->source, eff);
	eff->sound_holder = NULL;
	/*never cull Sound2d*/
	eff->trav_flags |= TF_DONT_CULL;
}
static Bool SND2D_GetChannelVolume(SFNode *node, Float *vol)
{
	Float volume = ((M_Sound2D *)node)->intensity;
	vol[0] = vol[1] = vol[2] = vol[3] = vol[4] = vol[5] = volume;
	return (volume==1.0f) ? 0 : 1;
}

static u8 SND2D_GetPriority(SFNode *node)
{
	return 255;
}

void R3D_InitSound2D(Render3D *sr, SFNode *node)
{
	Sound2DStack *snd;
	SAFEALLOC(snd, sizeof(Sound2DStack));
	snd->snd_ifce.GetPriority = SND2D_GetPriority;
	snd->snd_ifce.GetChannelVolume = SND2D_GetChannelVolume;
	snd->snd_ifce.owner = node;
	Node_SetPrivate(node, snd);
	Node_SetRenderFunction(node, RenderSound2D);
	Node_SetPreDestroyFunction(node, DestroySound2D);
}


static Float snd_compute_gain(Float min_b, Float min_f, Float max_b, Float max_f, SFVec3f pos)
{	
	Float sqpos_x, sqpos_z;
	Float y_pos, x_pos, dist_ellip, viewp_dist, dist_from_foci_min, dist_from_foci_max, d_min, d_max, sqb_min, sqb_max;
	Float a_in = (min_f+min_b)/2;
	Float b_in = (Float) sqrt(min_b*min_f);
	Float alpha_min = (min_f-min_b)/2;
	Float dist_foci_min = (min_f-min_b);
	Float a_out = (max_f+max_b)/2;	//first ellipse axis
	Float b_out = (Float) sqrt(max_b*max_f);
	Float alpha_max = (max_f-max_b)/2; //origo from focus
	Float dist_foci_max = (max_f-max_b);
	Float x_min = 0.0;
	Float x_max = 0.0;
	Float y_min = 0.0;
	Float y_max = 0.0;
	Float k = (fabs(pos.z) >= 0.00001) ? pos.x/pos.z : 0;

	sqpos_x = pos.x*pos.x;
	sqpos_z = pos.z*pos.z;

	dist_from_foci_min = (Float) (sqrt(sqpos_z + sqpos_x) + sqrt( (pos.z-dist_foci_min)*(pos.z-dist_foci_min) + sqpos_x) );
	dist_from_foci_max = (Float) (sqrt(sqpos_z + sqpos_x) + sqrt( (pos.z-dist_foci_max)*(pos.z-dist_foci_max) + sqpos_x) );	
	d_min = min_f+min_b;
	d_max = max_f+max_b;
	if(dist_from_foci_max > d_max) return 0.0f;
	else if (dist_from_foci_min <= d_min) return 1.0f;

	sqb_min = b_in*b_in;
	sqb_max = b_out*b_out;

	if(pos.z > 0.0001) {
		x_min = (Float) ( (alpha_min*sqb_min + a_in*b_in* sqrt( (a_in*k)*(a_in*k) + sqb_min - (alpha_min*k)*(alpha_min*k))) / (sqb_min + (a_in*k)*(a_in*k)) );
		y_min = k*x_min;
		x_max = (Float) ( (alpha_max*sqb_max + a_out*b_out* sqrt( (a_out*k)*(a_out*k) + sqb_max - (alpha_max*k)*(alpha_max*k))) / (sqb_max + (a_out*k)*(a_out*k)) );
		y_max = k*x_max;
	} else if (pos.z < -0.0001) {
		x_min = (Float) ( (alpha_min*sqb_min - a_in*b_in* sqrt( (a_in*k)*(a_in*k) + sqb_min - (alpha_min*k)*(alpha_min*k))) / (sqb_min + (a_in*k)*(a_in*k)) );
		y_min = k*x_min;
		x_max = (Float) ( (alpha_max*sqb_max - a_out*b_out* sqrt( (a_out*k)*(a_out*k) + sqb_max - (alpha_max*k)*(alpha_max*k))) / (sqb_max + (a_out*k)*(a_out*k)) );
		y_max = k*x_max;
	} else {
		x_min = x_max = 0;
		y_min = (Float) (b_in*sqrt(1 - (alpha_min/a_in)*(alpha_min/a_in)) );
		y_max = (Float) (b_out*sqrt(1 - (alpha_max/a_out)*(alpha_max/a_out)) );
	}

	y_pos = (Float) sqrt(sqpos_x) - y_min;
	x_pos = pos.z - x_min;
	x_max -= x_min;
	y_max -= y_min;
	dist_ellip = (Float) sqrt( y_max*y_max + x_max*x_max);
	viewp_dist = (Float) sqrt( y_pos*y_pos + x_pos*x_pos);
	return (Float) pow(10.0,(-(viewp_dist/dist_ellip)));
}


typedef struct
{
	SoundInterface snd_ifce;
	M4Matrix mx;
	SFVec3f last_pos;
	Bool identity;
	/*local system*/
	Float intensity;
	Float lgain, rgain;
} SoundStack;

static void DestroySound(SFNode *node)
{
	SoundStack *st = (SoundStack *)Node_GetPrivate(node);
	free(st);
}

static void RenderSound(SFNode *node, void *rs)
{
	RenderEffect3D *eff = (RenderEffect3D*) rs;
	M_Sound *snd = (M_Sound *)node;
	SoundStack *st = (SoundStack *)Node_GetPrivate(node);

	if (!snd->source) return;

	eff->sound_holder = &st->snd_ifce;

	/*forward in case we're switched off*/
	if (eff->trav_flags & TF_SWITCHED_OFF) {
		Node_Render((SFNode *) snd->source, eff);
	}
	else if (eff->traversing_mode==TRAVERSE_GET_BOUNDS) {
		/*we can't cull sound since*/
		eff->trav_flags |= TF_DONT_CULL;
	} else if (eff->traversing_mode==TRAVERSE_SORT) {
		M4Matrix mx;
		SFVec3f usr, snd_dir, pos;
		Float mag, ang;
		/*this implies no DEF/USE for real location...*/
		mx_copy(st->mx, eff->model_matrix);
		mx_copy(mx, eff->model_matrix);
		mx_inverse(&mx);

		snd_dir = snd->direction;
		vec_norm(snd_dir);

		/*get user location*/
		usr = eff->camera->position;
		mx_apply_vec(&mx, &usr);

		/*recenter to ellipse focal*/
		vec_diff(usr, usr, snd->location);
		mag = vec_len(usr);
		if (!mag) mag = 0.1f;
		ang = vec_dot(snd_dir, usr) / mag;
	
		usr.z = ang*mag;
		usr.x = (Float) sqrt(mag*mag - usr.z * usr.z);
		usr.y = 0;
		if (!vec_equal(usr, st->last_pos)) {
			st->intensity = snd_compute_gain(snd->minBack, snd->minFront, snd->maxBack, snd->maxFront, usr);
			st->intensity *= snd->intensity;
			st->last_pos = usr;
		}
		st->identity = (st->intensity==1.0f) ? 1 : 0;

		if (snd->spatialize) {
			Float ang, sign;
			SFVec3f cross;
			pos = snd->location;
			mx_apply_vec(&eff->model_matrix, &pos);
			vec_diff(pos, pos, eff->camera->position);
			vec_diff(usr, eff->camera->target, eff->camera->position);
			vec_norm(pos);
			vec_norm(usr);

			ang = (Float) acos(vec_dot(usr, pos));
			/*get orientation*/
			vec_cross(cross, usr, pos);
			sign = vec_dot(cross, eff->camera->up);
			if (sign>0) ang *= -1;
			ang = (1 + (Float) sin(ang)) / 2;
			st->lgain = (1 - ang*ang);
			st->rgain = 1 - (1 - ang)*(1 - ang);
			/*renorm between 0 and 1*/
			st->lgain *= 4*st->intensity/3;
			st->rgain *= 4*st->intensity/3;

			if (st->identity && ((st->lgain!=1.0f) || (st->rgain!=1.0f))) st->identity = 0;
		} else {
			st->lgain = st->rgain = 1.0f;
		}
		Node_Render((SFNode *) snd->source, eff);
	}

	eff->sound_holder = NULL;
}
static Bool SND_GetChannelVolume(SFNode *node, Float *vol)
{
	M_Sound *snd = (M_Sound *)node;
	SoundStack *st = (SoundStack *)Node_GetPrivate(node);

	vol[2] = vol[3] = vol[4] = vol[5] = st->intensity;
	if (snd->spatialize) {
		vol[0] = st->lgain;
		vol[1] = st->rgain;
	} else {
		vol[0] = vol[1] = st->intensity;
	}
	return !st->identity;
}
static u8 SND_GetPriority(SFNode *node)
{
	return (u8) ((M_Sound *)node)->priority*255;
}

void R3D_InitSound(Render3D *sr, SFNode *node)
{
	SoundStack *snd;
	SAFEALLOC(snd, sizeof(SoundStack));
	snd->snd_ifce.GetChannelVolume = SND_GetChannelVolume;
	snd->snd_ifce.GetPriority = SND_GetPriority;
	snd->snd_ifce.owner = node;
	Node_SetPrivate(node, snd);
	Node_SetRenderFunction(node, RenderSound);
	Node_SetPreDestroyFunction(node, DestroySound);
}

